# fmt: off
# Generated by the protocol buffer compiler.  DO NOT EDIT!
from __future__ import annotations
from datetime import datetime, timedelta
import enum
from typing import Any, Dict, List, Optional, TypeVar, Type
from google.protobuf import timestamp_pb2
from google.protobuf import duration_pb2
from google.protobuf.json_format import MessageToDict

from continual.rpc.logging.v1 import logging_pb2

from continual.rpc.management.v1 import types as management_v1
from google import protobuf

T = TypeVar("T")

def _to_proto_dict(x: dict) -> dict:
    """Converts dict to proto friendly representation."""
    out = dict()
    for key, item in x.items():
        if isinstance(item, list):
            out_list = []
            for e in item:
                if isinstance(e, datetime):
                    dt = timestamp_pb2.Timestamp()
                    dt.FromDatetime(e)
                    out_list.append(dt)
                elif isinstance(e, timedelta):
                    td = duration_pb2.Duration()
                    td.FromTimedelta(e)
                    out_list.append(td)
                elif hasattr(e, "to_proto"):
                    out_list.append(e.to_proto())
                else:
                    out_list.append(e)
            out[key] = out_list
        elif isinstance(item, datetime):
            dt = timestamp_pb2.Timestamp()
            dt.FromDatetime(item)
            out[key] = dt
        elif isinstance(item, timedelta):
            td = duration_pb2.Duration()
            td.FromTimedelta(item)
            out[key] = td
        elif hasattr(item, "to_proto"):
            out[key] = item.to_proto()
        else:
            out[key] = item
    return out


def _set_attribute(
    attributes: dict, class_name: str, field: str, value: Any, fieldType: Any, oneOfField: str
):
    """Helper that sets attribute on the object.
    
    Accepts dicts for classes and normalizes enums.
    """
    if isinstance(value, dict) and getattr(fieldType, "__origin__", None) == dict:
        attributes[field] = value
    elif isinstance(value, dict):
        attributes[field] = fieldType(**value)
    elif isinstance(value, list):
        innerType = fieldType.__args__[0]
        if innerType in [str, float, bool]:
            attributes[field] = value
        elif innerType is int:
            attributes[field] = [int(x) for x in value]
        elif issubclass(innerType, Enum):
            if len(value) > 0:
                if isinstance(value[0], str):
                    attributes[field] = [innerType(x.upper()) for x in value]
                elif isinstance(value[0], int) and getattr(innerType, "from_proto", None) is not None:
                    attributes[field] = [innerType.from_proto(x) for x in value]
        else:
            attributes[field] = [innerType(**x) for x in value]
    elif isinstance(value, str) and issubclass(fieldType, Enum):
        attributes[field] = fieldType(value.upper())
    elif isinstance(value, int) and issubclass(fieldType, Enum) and getattr(fieldType, "from_proto", None) is not None:
        attributes[field] = fieldType.from_proto(value)
    elif isinstance(value, str) and fieldType == datetime:
        dt = timestamp_pb2.Timestamp()
        dt.FromJsonString(value)
        attributes[field] = dt.ToDatetime()
    elif isinstance(value, str) and fieldType == timedelta:
        td = timestamp_pb2.Duration()
        td.FromJsonString(value)
        attributes[field] = td.ToTimeDelta()
    elif isinstance(value, str) and fieldType == int:
        # Needed to handle int64 encoded as strings.
        attributes[field] = int(value)
    # Set zero values to be consistent with proto.
    elif value is None:
        if fieldType == str:
            attributes[field] = ""
        elif fieldType == int:
            attributes[field] = 0
        elif fieldType == float:
            attributes[field] = 0.0
        elif fieldType == bool:
            attributes[field] = False
        elif getattr(fieldType, "__origin__", None) == list:
            attributes[field] = []
        elif getattr(fieldType, "__origin__", None) == dict:
            attributes[field] = {}
        elif fieldType == datetime or fieldType == timedelta:
            attributes[field] = None
        else:
            if oneOfField == "":
                attributes[field] = fieldType()
            else:
                attributes[field] = None
    elif value is not None and not isinstance(value, fieldType):
        raise ValueError(
            f"'{class_name}.{field}' must have type {fieldType}, got {type(value)}"
        )
    else:
        attributes[field] = value


class Enum(enum.Enum):
    """Base enum class.
    
    Overrides equality to be case insentitive.
    """

    def __eq__(self, other) -> bool:
        """Object is equal with all types."""
        if isinstance(other, str):
            return self.name == other.upper()
        elif isinstance(other, int):
            return self.value == other
        else:
            return self.value == other.value
    
    def __str__(self) -> str:
        return self.value




class LogHeartbeatRequest:
    """Log heartbeat request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , resource: str = ""
    , **kwargs) -> LogHeartbeatRequest:
        self._attributes = dict()
        self.resource = resource

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def resource(self) -> str:
        """Resource."""
        return self._attributes["resource"]


    @resource.setter
    def resource(self, value: str) -> None:
        _set_attribute(self._attributes, "LogHeartbeatRequest", "resource", value, str, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogHeartbeatRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogHeartbeatRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogHeartbeatRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogHeartbeatRequest object " + repr(self.to_dict()) + ">"


class LogExperimentsRequest:
    """Log experiments request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , model_version: str = ""
        , experiments: List[management_v1.Experiment] = []
    , **kwargs) -> LogExperimentsRequest:
        self._attributes = dict()
        self.model_version = model_version
        self.experiments = experiments

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def model_version(self) -> str:
        """Model version."""
        return self._attributes["model_version"]


    @model_version.setter
    def model_version(self, value: str) -> None:
        _set_attribute(self._attributes, "LogExperimentsRequest", "model_version", value, str, "")


    @property
    def experiments(self) -> List[management_v1.Experiment]:
        """Experiments."""
        return self._attributes["experiments"]


    @experiments.setter
    def experiments(self, value: List[management_v1.Experiment]) -> None:
        _set_attribute(self._attributes, "LogExperimentsRequest", "experiments", value, List[management_v1.Experiment], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogExperimentsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogExperimentsRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogExperimentsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogExperimentsRequest object " + repr(self.to_dict()) + ">"


class LogMetricsRequest:
    """Log metrics request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , model_version: str = ""
        , metrics: List[management_v1.TrainingMetrics] = []
    , **kwargs) -> LogMetricsRequest:
        self._attributes = dict()
        self.model_version = model_version
        self.metrics = metrics

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def model_version(self) -> str:
        """Model version."""
        return self._attributes["model_version"]


    @model_version.setter
    def model_version(self, value: str) -> None:
        _set_attribute(self._attributes, "LogMetricsRequest", "model_version", value, str, "")


    @property
    def metrics(self) -> List[management_v1.TrainingMetrics]:
        """Metrics."""
        return self._attributes["metrics"]


    @metrics.setter
    def metrics(self, value: List[management_v1.TrainingMetrics]) -> None:
        _set_attribute(self._attributes, "LogMetricsRequest", "metrics", value, List[management_v1.TrainingMetrics], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogMetricsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogMetricsRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogMetricsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogMetricsRequest object " + repr(self.to_dict()) + ">"


class LogEvaluationModelMetricsRequest:
    """Log evaluation model metrics request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , model_version: str = ""
        , evaluation_model_version_metrics: management_v1.EvaluationModelVersionMetrics = None
    , **kwargs) -> LogEvaluationModelMetricsRequest:
        self._attributes = dict()
        self.model_version = model_version
        self.evaluation_model_version_metrics = evaluation_model_version_metrics

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def model_version(self) -> str:
        """Model version."""
        return self._attributes["model_version"]


    @model_version.setter
    def model_version(self, value: str) -> None:
        _set_attribute(self._attributes, "LogEvaluationModelMetricsRequest", "model_version", value, str, "")


    @property
    def evaluation_model_version_metrics(self) -> management_v1.EvaluationModelVersionMetrics:
        """Evaluation model version metrics."""
        return self._attributes["evaluation_model_version_metrics"]


    @evaluation_model_version_metrics.setter
    def evaluation_model_version_metrics(self, value: management_v1.EvaluationModelVersionMetrics) -> None:
        _set_attribute(self._attributes, "LogEvaluationModelMetricsRequest", "evaluation_model_version_metrics", value, management_v1.EvaluationModelVersionMetrics, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogEvaluationModelMetricsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogEvaluationModelMetricsRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogEvaluationModelMetricsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogEvaluationModelMetricsRequest object " + repr(self.to_dict()) + ">"


class LogDatasetStatsRequest:
    """Log dataset stats request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , resource: str = ""
        , dataset_stats: management_v1.DatasetStats = None
    , **kwargs) -> LogDatasetStatsRequest:
        self._attributes = dict()
        self.resource = resource
        self.dataset_stats = dataset_stats

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def resource(self) -> str:
        """Resource."""
        return self._attributes["resource"]


    @resource.setter
    def resource(self, value: str) -> None:
        _set_attribute(self._attributes, "LogDatasetStatsRequest", "resource", value, str, "")


    @property
    def dataset_stats(self) -> management_v1.DatasetStats:
        """Dataset stats."""
        return self._attributes["dataset_stats"]


    @dataset_stats.setter
    def dataset_stats(self, value: management_v1.DatasetStats) -> None:
        _set_attribute(self._attributes, "LogDatasetStatsRequest", "dataset_stats", value, management_v1.DatasetStats, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogDatasetStatsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogDatasetStatsRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogDatasetStatsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogDatasetStatsRequest object " + repr(self.to_dict()) + ">"


class LogPredictionStatsRequest:
    """Log prediction stats request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , batch_prediction: str = ""
        , prediction_stats: management_v1.DatasetStats = None
    , **kwargs) -> LogPredictionStatsRequest:
        self._attributes = dict()
        self.batch_prediction = batch_prediction
        self.prediction_stats = prediction_stats

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def batch_prediction(self) -> str:
        """Batch prediction."""
        return self._attributes["batch_prediction"]


    @batch_prediction.setter
    def batch_prediction(self, value: str) -> None:
        _set_attribute(self._attributes, "LogPredictionStatsRequest", "batch_prediction", value, str, "")


    @property
    def prediction_stats(self) -> management_v1.DatasetStats:
        """Prediction stats."""
        return self._attributes["prediction_stats"]


    @prediction_stats.setter
    def prediction_stats(self, value: management_v1.DatasetStats) -> None:
        _set_attribute(self._attributes, "LogPredictionStatsRequest", "prediction_stats", value, management_v1.DatasetStats, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogPredictionStatsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogPredictionStatsRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogPredictionStatsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogPredictionStatsRequest object " + repr(self.to_dict()) + ">"


class GenerateArtifactUploadURLRequest:
    """Generate artifact upload u r l request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , resource: str = ""
        , path: str = ""
        , type: str = ""
        , mime_type: str = ""
        , metadata: str = ""
        , resumable: bool = False
    , **kwargs) -> GenerateArtifactUploadURLRequest:
        self._attributes = dict()
        self.resource = resource
        self.path = path
        self.type = type
        self.mime_type = mime_type
        self.metadata = metadata
        self.resumable = resumable

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def resource(self) -> str:
        """Resource."""
        return self._attributes["resource"]


    @resource.setter
    def resource(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "resource", value, str, "")


    @property
    def path(self) -> str:
        """Path."""
        return self._attributes["path"]


    @path.setter
    def path(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "path", value, str, "")


    @property
    def type(self) -> str:
        """Type."""
        return self._attributes["type"]


    @type.setter
    def type(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "type", value, str, "")


    @property
    def mime_type(self) -> str:
        """Mime type."""
        return self._attributes["mime_type"]


    @mime_type.setter
    def mime_type(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "mime_type", value, str, "")


    @property
    def metadata(self) -> str:
        """Metadata."""
        return self._attributes["metadata"]


    @metadata.setter
    def metadata(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "metadata", value, str, "")


    @property
    def resumable(self) -> bool:
        """Resumable."""
        return self._attributes["resumable"]


    @resumable.setter
    def resumable(self, value: bool) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLRequest", "resumable", value, bool, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.GenerateArtifactUploadURLRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.GenerateArtifactUploadURLRequest:
        """Convert object to proto representation."""
        return logging_pb2.GenerateArtifactUploadURLRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<GenerateArtifactUploadURLRequest object " + repr(self.to_dict()) + ">"


class GenerateArtifactUploadURLResponse:
    """Generate artifact upload u r l response."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , url: str = ""
        , artifact: management_v1.Artifact = None
    , **kwargs) -> GenerateArtifactUploadURLResponse:
        self._attributes = dict()
        self.url = url
        self.artifact = artifact

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def url(self) -> str:
        """Url."""
        return self._attributes["url"]


    @url.setter
    def url(self, value: str) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLResponse", "url", value, str, "")


    @property
    def artifact(self) -> management_v1.Artifact:
        """Artifact."""
        return self._attributes["artifact"]


    @artifact.setter
    def artifact(self, value: management_v1.Artifact) -> None:
        _set_attribute(self._attributes, "GenerateArtifactUploadURLResponse", "artifact", value, management_v1.Artifact, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.GenerateArtifactUploadURLResponse, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.GenerateArtifactUploadURLResponse:
        """Convert object to proto representation."""
        return logging_pb2.GenerateArtifactUploadURLResponse(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<GenerateArtifactUploadURLResponse object " + repr(self.to_dict()) + ">"


class LogDataChecksRequest:
    """Log data checks request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , resource: str = ""
        , data_checks: management_v1.DataChecks = None
    , **kwargs) -> LogDataChecksRequest:
        self._attributes = dict()
        self.resource = resource
        self.data_checks = data_checks

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def resource(self) -> str:
        """Resource."""
        return self._attributes["resource"]


    @resource.setter
    def resource(self, value: str) -> None:
        _set_attribute(self._attributes, "LogDataChecksRequest", "resource", value, str, "")


    @property
    def data_checks(self) -> management_v1.DataChecks:
        """Data checks."""
        return self._attributes["data_checks"]


    @data_checks.setter
    def data_checks(self, value: management_v1.DataChecks) -> None:
        _set_attribute(self._attributes, "LogDataChecksRequest", "data_checks", value, management_v1.DataChecks, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogDataChecksRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogDataChecksRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogDataChecksRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogDataChecksRequest object " + repr(self.to_dict()) + ">"


class LogModelVersionRequest:
    """Log model version request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , model_version: management_v1.ModelVersion = None
        , update_paths: List[str] = []
    , **kwargs) -> LogModelVersionRequest:
        self._attributes = dict()
        self.model_version = model_version
        self.update_paths = update_paths

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def model_version(self) -> management_v1.ModelVersion:
        """Model version."""
        return self._attributes["model_version"]


    @model_version.setter
    def model_version(self, value: management_v1.ModelVersion) -> None:
        _set_attribute(self._attributes, "LogModelVersionRequest", "model_version", value, management_v1.ModelVersion, "")


    @property
    def update_paths(self) -> List[str]:
        """Update paths."""
        return self._attributes["update_paths"]


    @update_paths.setter
    def update_paths(self, value: List[str]) -> None:
        _set_attribute(self._attributes, "LogModelVersionRequest", "update_paths", value, List[str], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogModelVersionRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogModelVersionRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogModelVersionRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogModelVersionRequest object " + repr(self.to_dict()) + ">"


class LogBatchPredictionRequest:
    """Log batch prediction request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , batch_prediction: management_v1.BatchPrediction = None
        , update_paths: List[str] = []
    , **kwargs) -> LogBatchPredictionRequest:
        self._attributes = dict()
        self.batch_prediction = batch_prediction
        self.update_paths = update_paths

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def batch_prediction(self) -> management_v1.BatchPrediction:
        """Batch prediction."""
        return self._attributes["batch_prediction"]


    @batch_prediction.setter
    def batch_prediction(self, value: management_v1.BatchPrediction) -> None:
        _set_attribute(self._attributes, "LogBatchPredictionRequest", "batch_prediction", value, management_v1.BatchPrediction, "")


    @property
    def update_paths(self) -> List[str]:
        """Update paths."""
        return self._attributes["update_paths"]


    @update_paths.setter
    def update_paths(self, value: List[str]) -> None:
        _set_attribute(self._attributes, "LogBatchPredictionRequest", "update_paths", value, List[str], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.LogBatchPredictionRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.LogBatchPredictionRequest:
        """Convert object to proto representation."""
        return logging_pb2.LogBatchPredictionRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<LogBatchPredictionRequest object " + repr(self.to_dict()) + ">"


class GetArtifactRequest:
    """Get artifact request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , name: str = ""
    , **kwargs) -> GetArtifactRequest:
        self._attributes = dict()
        self.name = name

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def name(self) -> str:
        """Name."""
        return self._attributes["name"]


    @name.setter
    def name(self, value: str) -> None:
        _set_attribute(self._attributes, "GetArtifactRequest", "name", value, str, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.GetArtifactRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.GetArtifactRequest:
        """Convert object to proto representation."""
        return logging_pb2.GetArtifactRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<GetArtifactRequest object " + repr(self.to_dict()) + ">"


class BatchGetArtifactsRequest:
    """Batch get artifacts request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , parent: str = ""
        , names: List[str] = []
    , **kwargs) -> BatchGetArtifactsRequest:
        self._attributes = dict()
        self.parent = parent
        self.names = names

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def parent(self) -> str:
        """Parent."""
        return self._attributes["parent"]


    @parent.setter
    def parent(self, value: str) -> None:
        _set_attribute(self._attributes, "BatchGetArtifactsRequest", "parent", value, str, "")


    @property
    def names(self) -> List[str]:
        """Names."""
        return self._attributes["names"]


    @names.setter
    def names(self, value: List[str]) -> None:
        _set_attribute(self._attributes, "BatchGetArtifactsRequest", "names", value, List[str], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.BatchGetArtifactsRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.BatchGetArtifactsRequest:
        """Convert object to proto representation."""
        return logging_pb2.BatchGetArtifactsRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<BatchGetArtifactsRequest object " + repr(self.to_dict()) + ">"


class BatchGetArtifactsResponse:
    """Batch get artifacts response."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , artifacts: List[management_v1.Artifact] = []
    , **kwargs) -> BatchGetArtifactsResponse:
        self._attributes = dict()
        self.artifacts = artifacts

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def artifacts(self) -> List[management_v1.Artifact]:
        """Artifacts."""
        return self._attributes["artifacts"]


    @artifacts.setter
    def artifacts(self, value: List[management_v1.Artifact]) -> None:
        _set_attribute(self._attributes, "BatchGetArtifactsResponse", "artifacts", value, List[management_v1.Artifact], "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.BatchGetArtifactsResponse, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.BatchGetArtifactsResponse:
        """Convert object to proto representation."""
        return logging_pb2.BatchGetArtifactsResponse(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<BatchGetArtifactsResponse object " + repr(self.to_dict()) + ">"


class DeleteArtifactRequest:
    """Delete artifact request."""

    _attributes: dict
    _client: Any = None
    
    def __init__(self
        , name: str = ""
    , **kwargs) -> DeleteArtifactRequest:
        self._attributes = dict()
        self.name = name

        if "client" in kwargs:
            self._client = kwargs["client"]
        if hasattr(self, "_init"):
            self._init()


    @property
    def name(self) -> str:
        """Name."""
        return self._attributes["name"]


    @name.setter
    def name(self, value: str) -> None:
        _set_attribute(self._attributes, "DeleteArtifactRequest", "name", value, str, "")


    @classmethod
    def from_proto(cls: Type[T], proto: logging_pb2.DeleteArtifactRequest, **kwargs) -> T:
        """Convert object from proto representation."""
        return cls(**MessageToDict(proto, preserving_proto_field_name=True), **kwargs)

    def to_proto(self) -> logging_pb2.DeleteArtifactRequest:
        """Convert object to proto representation."""
        return logging_pb2.DeleteArtifactRequest(**_to_proto_dict(self._attributes))

    def to_dict(self) -> dict:
        """Convert object to plain python dictionary representation."""
        return MessageToDict(
            self.to_proto(), 
            including_default_value_fields=True, 
            preserving_proto_field_name=True,
        )
    
    def __repr__(self) -> str:
        return "<DeleteArtifactRequest object " + repr(self.to_dict()) + ">"



import os
import re
from pathlib import Path
from configparser import ConfigParser


class FeaturestoreConfig:
    """Feature Store Config.

    Reads configuration from `~/.continual/.credentials` or `CONTINUAL_FS_CREDENTIALS` if
    not set during initialization.  Priority of config variables is `__init__`,
    environmental variables, and then the configuration file.

    Environmental variables are:

    * `CONTINUAL_FS_CREDENTIALS` - Credentials file.
    """

    _fsconfigfile: str

    _cp: ConfigParser

    def __init__(self) -> None:
        """Initialize featurestore config."""
        self._fsconfigfile = os.environ.get(
            "CONTINUAL_FS_CREDENTIALS",
            os.path.join(Path.home(), ".continual", ".credentials"),
        )

        Path(os.path.dirname(self._fsconfigfile)).mkdir(parents=True, exist_ok=True)

        if not Path(self._fsconfigfile).exists():
            Path(self._fsconfigfile).touch()

        self._cp = ConfigParser()
        self._cp.optionxform = str
        self._cp.read([self._fsconfigfile])

    def list(self, details=False, unmask=False):
        """Prints all featurestore config."""
        print("Reading File: %s" % self._fsconfigfile)
        if details:
            with open(self._fsconfigfile, "r") as f:
                text = f.read()
                if not unmask:
                    text = re.sub(
                        "(password\\s*=\\s*.+)",
                        "PASSWORD = XXXXXXXXXXXXXXXXXXX",
                        text,
                        flags=re.IGNORECASE,
                    )
                print(text)
                f.close()
        else:
            print("Sections: ")
            for s in self._cp.sections():
                print(s)

    def get(self, name):
        """Prints a single featurestore config."""
        print("[%s]" % name)
        for key in self._cp[name].keys():
            if key.lower() == "password":
                val = "XXXXXXXXXXXXXXXXXXX"
            else:
                val = self._cp[name][key]
            print("%s = %s" % (key, val))

    def create(self, name, dict):
        """Creates featurestore config"""
        # force keys to be lowercase
        dict = {k.lower(): v for k, v in dict.items()}
        self._cp[name] = dict

        with open(self._fsconfigfile, "w") as f:
            self._cp.write(f)

    def delete(self, name):
        """Deletes featurestore config."""
        self._cp.remove_section(name)
        with open(self._fsconfigfile, "w") as f:
            self._cp.write(f)

    def update(self, name, key, value):
        """Updates value in featurestore config."""
        if not self._cp.has_section(name):
            self._cp.add_section(name)
        self._cp[name][key.lower()] = value
        with open(self._fsconfigfile, "w") as f:
            self._cp.write(f)



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from google.protobuf import field_mask_pb2
from continual.python.sdk.events import EventManager


class UserManager(Manager):
    """Manages user resources."""

    name_pattern: str = "users/{user}"

    def get(self, id: str) -> User:
        """Get user.

        Arguments:
            id: User name or id.

        Returns
            A user.
        """
        req = management_pb2.GetUserRequest(name=self.name(id))
        resp = self.client._management.GetUser(req)
        return User.from_proto(resp, client=self.client)

    def list(self, page_size: Optional[int] = None) -> List[User]:
        """List users.

        Arguments:
            page_size: Number of itmes to return.

        Returns:
            A list of users.
        """
        req = management_pb2.ListUsersRequest(page_size=page_size)
        resp = self.client._management.ListUsers(req)
        return [User.from_proto(u, client=self.client) for u in resp.users]

    def list_all(self) -> Pager[User]:
        """List all users.

        Pages through all users using an iterator.

        Returns:
            A iterator of all users.
        """

        def next_page(next_page_token):
            req = management_pb2.ListUsersRequest(page_token=next_page_token)
            resp = self.client._management.ListUsers(req)
            return (
                [User.from_proto(u, client=self.client) for u in resp.users],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete a user.

        Arguments:
            id: Name or id.
        """

        req = management_pb2.DeleteUserRequest(name=self.name(id))
        self.client._management.DeleteUser(req)

    def update(
        self,
        id: str,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        bio: Optional[str] = None,
        location: Optional[str] = None,
    ) -> User:
        """Update user.

        Arguments:
            first_name:  First name of display name.
            last_name:  Last name of display name.
            bio: Bio.
            location: Location.

        Returns:
            Updated user.
        """
        paths = []
        if first_name is not None:
            paths.append("first_name")
        if last_name is not None:
            paths.append("last_name")
        if bio is not None:
            paths.append("bio")
        if location is not None:
            paths.append("location")
        req = management_pb2.UpdateUserRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            user=User(
                name=self.name(id),
                first_name=first_name,
                last_name=last_name,
                bio=bio,
                location=location,
            ).to_proto(),
        )
        resp = self.client._management.UpdateUser(req)
        return User.from_proto(resp, client=self.client)


class User(Resource, types.User):
    """User resource."""

    manager: UserManager
    events: EventManager
    name_pattern = "users/{users}"

    def _init(self):
        self.manager = UserManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def delete(self) -> None:
        """Delete user."""
        self.manager.delete(self.name)

    def update(
        self,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        bio: Optional[str] = None,
        location: Optional[str] = None,
    ) -> User:
        """Update user.

        Arguments:
            first_name:  First name of display name.
            last_name:  Last name of display name.
            bio: Bio.
            location: Location.

        Returns:
            Updated user.
        """
        return self.manager.update(
            self.name,
            first_name=first_name,
            last_name=last_name,
            bio=bio,
            location=location,
        )



from jinja2 import Template


model_prediction_classification_template = Template(
    """
        CREATE OR REPLACE VIEW {{params.DbSchema}}.{{params.ModelView}} AS
        SELECT entity_table.*, C.{{ params.TargetName}}_prediction, C.{{ params.TargetName}}_prediction_score, C.{{ params.TargetName}}_prediction_details from {{ params.ModelDataView}} AS entity_table
        LEFT JOIN
        (SELECT B.{{params.IndexColumn}}, {% if params.TimestampColumn %} B.{{params.TimestampColumn}}, {% endif %} B.{{ params.TargetName}}_prediction, B.{{ params.TargetName}}_prediction_score, B.{{ params.TargetName}}_prediction_details 
            FROM 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} max(prediction_time) AS p_time
                FROM {{params.DbSchema}}.{{params.ModelPredictionTable}}
                GROUP BY {{params.IndexColumn}} 
                {% if params.TimestampColumn %}, {{params.TimestampColumn}} {% endif %}) AS A
            LEFT JOIN 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} {{ params.TargetName}}_prediction, {{ params.TargetName}}_prediction_score, {{ params.TargetName}}_prediction_details, prediction_time 
                FROM  {{params.DbSchema}}.{{params.ModelPredictionTable}}) AS B
            ON A.{{params.IndexColumn}} = B.{{params.IndexColumn}} 
            {% if params.TimestampColumn %} AND A.{{params.TimestampColumn}}  = B.{{params.TimestampColumn}} {% endif %}
            AND A.p_time = B.prediction_time) AS C
        ON entity_table.{{params.IndexMatchName}} = C.{{params.IndexColumn}}  {% if params.TimestampColumn %}  AND entity_table.{{params.TSMatchName}} = C.{{params.TimestampColumn}}  {% endif %}
"""
)

model_prediction_regression_template = Template(
    """
        CREATE OR REPLACE VIEW {{params.DbSchema}}.{{params.ModelView}} AS
        SELECT entity_table.*, C.{{ params.TargetName}}_prediction from {{ params.ModelDataView}} AS entity_table
        LEFT JOIN
        (SELECT B.{{params.IndexColumn}}, {% if params.TimestampColumn %} B.{{params.TimestampColumn}}, {% endif %} B.{{ params.TargetName}}_prediction 
            FROM 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} max(prediction_time) AS p_time
                FROM {{params.DbSchema}}.{{params.ModelPredictionTable}}
                GROUP BY {{params.IndexColumn}} 
                {% if params.TimestampColumn %}, {{params.TimestampColumn}} {% endif %}) AS A
            LEFT JOIN 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} {{ params.TargetName}}_prediction, prediction_time 
                FROM  {{params.DbSchema}}.{{params.ModelPredictionTable}}) AS B
            ON A.{{params.IndexColumn}} = B.{{params.IndexColumn}} 
            {% if params.TimestampColumn %} AND A.{{params.TimestampColumn}}  = B.{{params.TimestampColumn}} {% endif %}
            AND A.p_time = B.prediction_time) AS C
        ON entity_table.{{params.IndexMatchName}} = C.{{params.IndexColumn}}  {% if params.TimestampColumn %}  AND entity_table.{{params.TSMatchName}} = C.{{params.TimestampColumn}}  {% endif %}
"""
)

model_prediction_binary_template = Template(
    """
        CREATE OR REPLACE VIEW {{params.DbSchema}}.{{params.ModelView}} AS
        SELECT entity_table.*, C.{{ params.TargetName}}_prediction, C.{{ params.TargetName}}_true_prediction_score from {{ params.ModelDataView}} AS entity_table
        LEFT JOIN
        (SELECT B.{{params.IndexColumn}}, {% if params.TimestampColumn %} B.{{params.TimestampColumn}}, {% endif %} B.{{ params.TargetName}}_prediction, B.{{ params.TargetName}}_true_prediction_score 
            FROM 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} max(prediction_time) AS p_time
                FROM {{params.DbSchema}}.{{params.ModelPredictionTable}}
                GROUP BY {{params.IndexColumn}} 
                {% if params.TimestampColumn %}, {{params.TimestampColumn}} {% endif %}) AS A
            LEFT JOIN 
                (SELECT {{params.IndexColumn}}, {% if params.TimestampColumn %} {{params.TimestampColumn}}, {% endif %} {{ params.TargetName}}_prediction, {{ params.TargetName}}_true_prediction_score, prediction_time 
                FROM  {{params.DbSchema}}.{{params.ModelPredictionTable}}) AS B
            ON A.{{params.IndexColumn}} = B.{{params.IndexColumn}} 
            {% if params.TimestampColumn %} AND A.{{params.TimestampColumn}}  = B.{{params.TimestampColumn}} {% endif %}
            AND A.p_time = B.prediction_time) AS C
        ON entity_table.{{params.IndexMatchName}} = C.{{params.IndexColumn}}  {% if params.TimestampColumn %}  AND entity_table.{{params.TSMatchName}} = C.{{params.TimestampColumn}}  {% endif %}
"""
)


def create_model_view_query(
    model_data_view,
    index_column,
    ts_column,
    index_match_name,
    ts_match_name,
    model_prediction_table,
    project,
    model_view_name,
    target_name,
    db_schema,
    problem_type,
):
    params = {
        "ModelDataView": model_data_view,
        "IndexColumn": index_column,
        "TimestampColumn": ts_column,
        "IndexMatchName": index_match_name,
        "TSMatchName": ts_match_name,
        "ModelPredictionTable": model_prediction_table,
        "ProjectName": project,
        "ModelView": model_view_name,
        "TargetName": target_name,
        "DbSchema": db_schema,
    }
    if problem_type == "regression" or problem_type == "forecasting":
        return model_prediction_regression_template.render(params=params)
    if problem_type == "multiclass_classification":
        return model_prediction_classification_template.render(params=params)
    if problem_type == "binary_classification":
        return model_prediction_binary_template.render(params=params)

    raise ValueError("Unable to handle problem type [{}]".format(problem_type))



import pandas as pd
import logging


class BaseCustomAlgorithm:
    """
    Base Class for all custom model implementations
    """

    def __init__(self, logger: logging.Logger, **kwargs):
        """
        Initialize any state

        Parameters
        ----------
        client: Client
            SDK client
        kwargs: dict
            Key word arguments
        """

        self.logger = logger

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        X_val: pd.DataFrame,
        y_val: pd.DataFrame,
        params: dict,
    ):
        """
        Construct and train model defined by config on data.
        Store any model state in self

        Parameters
        ----------
        X: DataFrame
            The training
        y: DataFrame
            The training outputs
        X_val: DataFrame
            validation features
        y_val: DataFrame
            validation outputs
        params: dict
            hyperparameters to the model
        """

        raise NotImplementedError

    def predict(self, X: pd.DataFrame, params: dict) -> pd.Series:
        """
        Run trained model on data. Utilize
        any model state in self.

        Note: Continual will ensure that fit()
        has been called before predict()

        Parameters
        ----------
        X : DataFrame
            The features from which to get predictions
        params: dict
            hyperparameters to the model

        Returns
        -------
        pandas.Series
            The class predictions for the input data
        """

        raise NotImplementedError

    def save(self, save_dir):
        """
        Serialize current instance state to disk in given
        directory (directory is already created)

        Parameters
        ----------
        save_dir : str
            Save path for serialized model file
        """

        raise NotImplementedError

    def load(self, load_dir: str):
        """
        Deserialize current instance state from given directory

        Parameters
        ----------
        load_dir: str
            The directory where the serialized files for this
            model are stored
        """
        raise NotImplementedError

    def default_parameters(self) -> dict:
        """
        Returns
        -------
        dict
            The default hyperparameters that will be
            passed to the fit function
        """

        raise NotImplementedError



import pandas as pd
import dill
import os
import logging
import sklearn.linear_model as sklm

from continual.python.sdk.extensions.algorithms.base_custom_algorithm import (
    BaseCustomAlgorithm,
)

"""
Template custom model class

ADD YOUR IMPLEMENTATION HERE.
"""


class MyCustomAlgorithm(BaseCustomAlgorithm):
    """
    My custom model
    """

    model_file_name: str = "custom_algo.pkl"

    def __init__(self, logger: logging.Logger, **kwargs):
        """
        Initialize any state

        Parameters
        ----------
        logger: Logger
            The output logger initialized with log
            files for this model
        kwargs: dict
            Key word arguments
        """

        self.logger = logger
        self.model_type = sklm.ElasticNet

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        X_val: pd.DataFrame,
        y_val: pd.DataFrame,
        params: dict,
    ):
        """
        Construct and train model defined by config on data.
        Store any model state in self

        Parameters
        ----------
        X: DataFrame
            The training
        y: DataFrame
            The training outputs
        X_val: DataFrame
            validation features
        y_val: DataFrame
            validation outputs
        params: dict
            hyperparameters to the model
        """

        if self.model_type:
            self.model = self.model_type(**params).fit(X, y)
        else:
            raise Exception("Model type not initialized")

    def predict(self, X: pd.DataFrame, params: dict) -> pd.Series:
        """
        Run trained model on data. Utilize
        any model state in self.

        Note: Continual will ensure that fit()
        has been called before predict()

        Parameters
        ----------
        X : DataFrame
            The features from which to get predictions
        params: dict
            hyperparameters to the model

        Returns
        -------
        pd.Series
            The class predictions for the input data.

            IMPORTANT: If the problem type is binary
            classsification, this function should return
            an array of probabilities for the positive
            class. If it is multiclass classification,
            this function should return a numpy array
            of shape (X.shape[0], num_classes)representing
            array of probabilities for each class.
        """

        return self.model.predict(X)

    def save(self, save_dir) -> str:
        """
        Serialize current instance state to disk in given
        directory (directory is already created)

        Parameters
        ----------
        save_dir : str
            Save path for serialized model file
        """

        model_file_path = os.path.join(save_dir, self.model_file_name)
        with open(model_file_path, "wb") as model_file:
            dill.dump(self.model, model_file)

    def load(self, load_dir: str):
        """
        Deserialize current instance state from given directory

        Parameters
        ----------
        load_dir: str
            The directory where the serialized files for this
            model are stored
        """

        model_file_path = os.path.join(load_dir, self.model_file_name)
        with open(model_file_path, "rb") as model_file:
            self.model = dill.load(model_file)

    def default_parameters(self):
        """
        Returns
        -------
        dict
            The default hyperparameters that will be
            passed to the fit function
        """

        return {"random_state": 0}



import logging
import sqlalchemy as db
from sqlalchemy.dialects.postgresql import VARCHAR
from sqlalchemy.exc import IntegrityError, ProgrammingError, NoSuchTableError
import urllib.parse

import traceback
import pandas as pd
import json
from datetime import datetime

from .query_templates import stats_template
from .featurestore import FeatureStore, Typemap
from .prediction_store import PredictionStore, chunk_it
from continual.rpc.management.v1 import management_types_pb2
from .model_templates import create_model_view_query
from sqlalchemy.sql.schema import MetaData


class RedshiftFeatureStore(FeatureStore, PredictionStore):
    def __init__(self, data_store, source, cfg):
        self.data_store = data_store
        config = data_store.redshift
        db_url = "postgresql://{}:{}@{}:{}/{}".format(
            config.username,
            urllib.parse.quote_plus(config.password),
            config.host,
            config.port,
            config.database,
        )

        self.engine = db.create_engine(db_url)
        self.connection = self.engine.connect()
        self.trans = None
        super(RedshiftFeatureStore, self).__init__()

    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """

        return super().create_feature_table(feature_set)

    def get_feature_table(self, feature_set):
        project, tablename = self.get_table_name(feature_set)
        tablename = "featureset_{}".format(tablename)
        try:
            tab = db.Table(
                tablename,
                db.MetaData(schema=self.db_schema),
                autoload=True,
                autoload_with=self.engine,
            )
            return tab
        except NoSuchTableError:
            print("No table found error")
            return None

    def set_db_schema(self, db_schema):
        self.db_schema = db_schema
        try:
            self.connection.execute("SET search_path TO {},public".format(db_schema))
        except:
            pass

    def cleanup(self):
        self.connection.execute("DROP SCHEMA IF EXISTS " + self.db_schema + " CASCADE")
        return

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def get_type(self) -> str:
        return "redshift"

    def is_hosted(self):
        return True

    def get_stats_query_template(self):
        return stats_template

    def get_database_names(self):
        databases = []
        result_proxy = self.connection.execute(
            "SELECT database_name as name FROM svv_redshift_databases ORDER BY name"
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            databases.append(d["name"])

        return databases

    def get_schema_names(self, database):
        schemas = []
        result_proxy = self.connection.execute(
            "SELECT schema_name as name FROM svv_redshift_schemas WHERE database_name = '{}' ORDER BY name".format(
                database
            )
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            if d["name"] == "INFORMATION_SCHEMA":
                continue
            schemas.append(d["name"])

        return schemas

    def get_table_names(self, database, schema, views=True):
        tables = []
        result_proxy = self.connection.execute(
            "SELECT table_name as name FROM svv_redshift_tables WHERE database_name = '{}' AND schema_name = '{}' ORDER BY name".format(
                database, schema
            )
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            tables.append(d["name"])

        if views:
            pass
            # result_proxy = self.connection.execute(
            #     "SHOW VIEWS IN SCHEMA {}.{}".format(database, schema)
            # )
            # for rowproxy in result_proxy:
            #     d = {}
            #     for col, val in rowproxy.items():
            #         d[col] = val
            #     tables.append(d["name"])
        return tables

    def get_column_names(self, database, schema, table):
        columns = []
        type_map = {}
        result_proxy = self.connection.execute(
            "SELECT column_name, data_type FROM svv_redshift_columns WHERE database_name = '{}' AND schema_name = '{}' AND table_name = '{}' ORDER BY column_name".format(
                database, schema, table
            )
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            columns.append(d["column_name"])
            if "data_type" in d:
                type_map[d["column_name"]] = d["data_type"]

        return columns, type_map

    def _create_prediction_model_table(
        self, model, model_pred_name, meta: MetaData, problem_type
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """

        index = None
        time_index = None
        target_column = model.schema.target
        target_column_dtype = db.TEXT
        for f in model.schema.columns:
            if (
                f.type == management_types_pb2.FieldType.INDEX
                or f.name == model.schema.index
            ):
                index = f
            elif (
                f.type == management_types_pb2.FieldType.TIME_INDEX
                or f.name == model.schema.time_index
            ):
                time_index = f
            elif f.name == target_column:
                target_column_dtype = Typemap[f.dtype]

        columns = []
        columns.append(db.Column("batch_prediction", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP(timezone=True)))
        columns.append(db.Column(f"features", VARCHAR(65535)))

        if index is not None:
            columns.append(db.Column(index.name, Typemap[index.dtype]))

        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.dtype]))
        columns.append(db.Column(f"{target_column}_prediction", target_column_dtype))
        if problem_type == "multiclass_classification":
            columns.append(
                db.Column(f"{target_column}_prediction_details", VARCHAR(65535))
            )
            columns.append(db.Column(f"{target_column}_prediction_score", db.Float))
        if problem_type == "binary_classification":
            columns.append(
                db.Column(f"{target_column}_true_prediction_score", db.Float)
            )

        db.Table(model_pred_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    raise

    def get_prediction_table(self, model):
        project_id, pred_table = self.get_model_prediction_table_name(model)
        print("Prediction table in " + self.db_schema)
        try:
            tab = db.Table(
                pred_table,
                db.MetaData(schema=self.db_schema),
                autoload=True,
                autoload_with=self.engine,
            )
            return tab
        except NoSuchTableError:
            # Todo perhaps not capture the error here.
            return None

    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.debug(
            "in write_prediction for [{}]:[{}]".format(
                job_name, job_start_time.strftime("%Y-%m-%d %H:%M:%S")
            )
        )
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )
        df["prediction_time"] = job_start_time.strftime("%Y-%m-%d %H:%M:%S")

        db_url = f"postgresql://{self.data_store.redshift.username}:{urllib.parse.quote_plus(self.data_store.redshift.password)}@{self.data_store.redshift.host}:{self.data_store.redshift.port}/{self.data_store.redshift.database}"
        self.engine = db.create_engine(db_url)
        self.connection = self.engine.connect()

        tab = self.get_prediction_table(model)
        if tab is None:
            print("Prediction table not found")
            logging.debug("Prediction table not found")
            return

        # set chained assignent to warn in case an underlying model has set to raise
        # TODO: determine if there's a better method of mapping the json values
        pd.set_option("mode.chained_assignment", "warn")
        trans = self.connection.begin()

        prediction_details_column = f"{output_name}_prediction_details"

        ch = chunk_it(0, df.shape[0])
        while ch is not None:
            df_slice = df[ch[0] : ch[1]]
            if prediction_details_column in df_slice.columns:
                df_slice[prediction_details_column] = df_slice[
                    prediction_details_column
                ].map(json.dumps)

            try:
                df_slice.to_sql(
                    pred_table,
                    self.connection,
                    schema=self.db_schema,
                    if_exists="append",
                    method="multi",
                    index=False,
                )
            except:
                trans.rollback()
                traceback.print_exc()
                logging.debug("write failed to execute - aborting")
                raise
            ch = chunk_it(ch[1], df.shape[0])

        trans.commit()

        count_query = f"SELECT COUNT(1) from {self.db_schema}.{pred_table} WHERE batch_prediction = '{job_name}'"
        result = self.connection.execute(count_query)
        for row in result:
            return row[0]
        return 0

    def create_view(self, feature_set):
        project, table_name = self.get_table_name(feature_set)

        # drop the view first if it exists
        self.connection.execute(
            "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, table_name)
        )

        query = feature_set.schema.query.strip().rstrip(";")
        self.connection.execute(
            "CREATE OR REPLACE VIEW {}.{} AS ({}) WITH NO SCHEMA BINDING".format(
                self.db_schema, table_name, query
            )
        )
        return

    def create_featureset_view(self, feature_set):
        project, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        query = feature_set.schema.query.strip().rstrip(";")
        self.connection.execute(
            "CREATE OR REPLACE VIEW {}.{} AS ({}) WITH NO SCHEMA BINDING".format(
                self.db_schema, table_name, query
            )
        )
        return

    def create_model_data_view(self, model):
        project, table_name = self.get_data_table_name(model.name)

        # drop the view first if it exists
        self.connection.execute(
            "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, table_name)
        )

        query = model.schema.query.strip().rstrip(";")
        self.connection.execute(
            "CREATE OR REPLACE VIEW {}.{} AS ({}) WITH NO SCHEMA BINDING".format(
                self.db_schema, table_name, query
            )
        )
        return

    def create_model_view(self, model, fs_map, metadata, problem_type):
        """
        Create or refresh model historical prediction tables and a view that
          joins training data and predictions.
        """

        project, entity_name = self.get_data_table_name(model.name)
        table_name = model.schema.table
        if table_name is None or table_name == "":
            table_name = f"{self.db_schema}.{entity_name}"

        index = model.schema.index or None
        time_index = model.schema.time_index or None
        for f in model.schema.columns:
            logging.info(f"create_model_view column: {f}")
            if f.type == management_types_pb2.FieldType.INDEX:
                index = f.name
            elif f.type == management_types_pb2.FieldType.TIME_INDEX:
                time_index = f.name

        _, model_pred_name = self.get_model_prediction_table_name(model)
        proj, model_view_name = self.get_model_view_name(model)

        # The training data query returns qulified column names. in order to join this column name we need
        # to pass the name to be expected out of the traning query.
        index_match_name = index
        time_match_name = None
        if time_index:
            time_match_name = time_index

        model_query = create_model_view_query(
            table_name,
            index,
            time_index,
            index_match_name,
            time_match_name,
            model_pred_name,
            proj,
            model_view_name,
            model.schema.target,
            self.db_schema,
            problem_type,
        )
        # add redshift WITH NO SCHEMA BINDING
        model_query_final = f"{model_query} WITH NO SCHEMA BINDING"
        logging.info(model_query_final)
        try:
            # drop view first
            self.connection.execute(
                "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, model_view_name)
            )
            self.connection.execute(model_query_final)
        except StopIteration:
            pass
        except:
            logging.info(traceback.format_exc()[:3000])
            raise

    def infer_schema(
        self,
        query,
        excludes=None,
        sample_row_count=100,
        allow_invalid_types=False,
        timeout_ms=0,
    ):
        if timeout_ms and timeout_ms > 0:
            config = self.data_store.redshift
            db_url = "postgresql://{}:{}@{}:{}/{}".format(
                config.username,
                urllib.parse.quote_plus(config.password),
                config.host,
                config.port,
                config.database,
            )
            self.engine = db.create_engine(
                db_url,
                connect_args={"options": "-c statement_timeout={}".format(timeout_ms)},
            )
            self.connection = self.engine.connect()

            res = super().infer_schema(
                query, excludes, sample_row_count, allow_invalid_types, timeout_ms
            )

            self.engine = db.create_engine(db_url)
            self.connection = self.engine.connect()

            return res
        else:
            return super().infer_schema(
                query, excludes, sample_row_count, allow_invalid_types, timeout_ms
            )



from typing import Optional, List
import typer
import os
from pathlib import Path
import yaml

from continual.python.cli import utils
from continual.python.sdk.client import Client
from continual.rpc.management.v1.types import Extension
from rich.console import Console

app = typer.Typer(help="Manage extensions.", hidden=True)


def format_extension_data(extension: Extension, zipped=False, all_projects=False):
    create_time = extension.create_time.replace(microsecond=0)
    update_time = extension.update_time.replace(microsecond=0)

    environment = extension.parent.split("/")[-1]
    data = [
        extension.extension_id,
        extension.name,
        environment,
        extension.extension_type,
        create_time,
        update_time,
        extension.class_name,
        extension.module_name,
        extension.package_url,
        extension.package_hash,
    ]
    headers = [
        "ID",
        "Name",
        "Environment",
        "Extension Type",
        "Created",
        "Updated",
        "Class Name",
        "Module Name",
        "Package URL",
        "Package Hash",
    ]
    if zipped:
        return tuple([x[0], x[1]] for x in (zip(headers, data)))
    if all_projects:
        data.insert(0, extension.parent.split("/")[1])
        headers.insert(0, "Project")
    return (data, headers)


# use callback to run  list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        extension_list(
            project=None,
            environment=None,
            n=30,
            all_projects=False,
            style=None,
        )


@app.command("init")
@utils.exit_on_error
def extension_init(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    id: str = typer.Option(..., "--id", help="Unique ID/label for the extension."),
    path: str = typer.Option(
        os.getcwd(), "--path", help="Directory in which to create the extension."
    ),
    ext_type: str = typer.Option(
        "model", "--type", help="Set the extension type to a custom model."
    ),
    class_name: str = typer.Option(
        None, "--class", help="Name of the implementation class"
    ),
    source_path: str = typer.Option(
        None,
        "--source",
        help="Implementation source code to be copied into extension directory.",
    ),
    dependencies: Optional[List[str]] = typer.Option(
        None, "--dependency", help="Add a pip package dependency to this extension"
    ),
):
    """Create a new extension template."""

    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if ext_type.lower() not in {"algorithm", "model"}:
        typer.secho(
            "Unsupported extension type: %s \n" % ext_type,
            fg="red",
        )

    # Call to ExtensionManager
    c.extensions.init(
        root=path,
        id=id,
        extension_type=ext_type,
        class_name=class_name,
        source_path=source_path,
        dependencies=list(dependencies),
    )

    typer.secho(
        "\nCreated extension '%s' of type %s at %s"
        % (id, ext_type, os.path.join(path, id)),
        fg="blue",
    )


@app.command("list")
@utils.exit_on_error
def extension_list(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List Extensions."""

    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    project_snippet = "project %s, environment %s" % (project, environment)
    if all_projects:
        project_snippet = "all accessible projects"
    data = []
    headers = []
    for e in c.extensions.list(n, all_projects=all_projects):
        (e_data, headers) = format_extension_data(e, all_projects=all_projects)
        data.append(e_data)
    typer.secho(
        "\nFound %s extensions in %s:" % (len(data), project_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    extension: str = typer.Argument(..., help="Extension ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Get extension details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    e = c.extensions.get(id=extension)

    console = Console()
    if to_json:
        console.print(e.to_dict())
    else:
        data = format_extension_data(e, zipped=True)
        typer.secho("\nRetrieving extension %s: \n" % (extension), fg="blue")
        utils.print_info(data)
        typer.secho("\nExtension schema:\n", fg="blue")
        console.print(yaml.safe_load(e.schema_text))


@app.command("export")
@utils.exit_on_error
def export(
    extension: str = typer.Argument(..., help="Extension ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    extract: bool = typer.Option(
        False,
        "--extract",
        help="Untar the dowloaded tarball containing the source code.",
    ),
    path: Path = typer.Option(
        None,
        help="Directory to save extension source code into. Defaults to ./<project_id> if not provided.",
    ),
):
    """Export extension."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if path is None:
        path = os.getcwd()

    export_name = c.extensions.export(id=extension, path=path, extract=extract)
    typer.secho(f"Saving extension {extension} to {export_name}", fg="blue")


@app.command("delete")
@utils.exit_on_error
def delete(
    extension: str = typer.Argument(..., help="Extension ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete extension."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if not force:
        force = typer.confirm(
            "Are you sure you want to delete extension %s in project %s, environment %s?"
            % (extension, project, environment),
        )
    if force:
        c.extensions.delete(id=extension)
        typer.secho("Successfully deleted extension %s." % extension, fg="green")
    else:
        typer.secho("Cancelled extension deletion!", fg="blue")



import typer
import json
import pandas as pd
import rich
import yaml

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console
from typing import List

app = typer.Typer(help="Manage model versions.")


def format_model_version_data(mv, zipped=False, all_projects=False):
    try:
        experiment = mv.experiment.id
    except:
        experiment = "N/A"
    if mv.promoted:
        last_promoted = Client().promotions.get(mv.promotion).promoted_time
        if last_promoted:
            last_promoted = last_promoted.replace(microsecond=0)
    else:
        last_promoted = "N/A"
    update_time = mv.update_time.replace(microsecond=0)
    create_time = mv.create_time.replace(microsecond=0)
    if zipped:
        data = [
            mv.id,
            mv.name,
            mv.state.value,
            experiment,
            mv.performance_metric,
            mv.performance_metric_val,
            mv.promoted,
            mv.promotion,
            last_promoted,
            create_time,
            update_time,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Winning Experiment",
            "Metric",
            "Metric Value",
            "Promoted",
            "Promotion",
            "Promoted Time",
            "Created",
            "Updated",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            mv.id,
            mv.state.value,
            mv.name.split("/")[3],
            experiment,
            mv.performance_metric,
            mv.performance_metric_val,
            mv.promoted,
            last_promoted,
            create_time,
            # update_time,
        ]
        headers = [
            "ID",
            "State",
            "Model",
            "Winning Experiment",
            "Metric",
            "Metric Value",
            "Promoted",
            "Promoted Time",
            "Created",
            # "Updated",
        ]
        if all_projects:
            data.insert(0, mv.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        mv_list(
            model=None,
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def mv_list(
    model: str = typer.Option(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List model versions.

    Filters can include:
        --state (i.e. state:FAILED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    model_snippet = ""
    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if model is not None:
        c = c.models.get(model)
        model_snippet = "for model %s " % model
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters)
    if all_projects:
        project_snippet = "all accessible projects"
    data = []
    headers = []
    for mv in c.model_versions.list(n, filters=filters, all_projects=all_projects):
        (mv_data, headers) = format_model_version_data(mv, all_projects=all_projects)
        data.append(mv_data)
    typer.secho(
        "\nFound %s model versions %sin  %s%s: "
        % (len(data), model_snippet, project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    model_version: str = typer.Argument(..., help="Model version ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Get model version information."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.model_versions.get(model_version)
    console = Console()
    if to_json:
        console.print(mv.to_dict())
    else:
        data = format_model_version_data(mv, zipped=True)
        typer.secho("\nRetrieving model version %s: \n" % (model_version), fg="blue")
        utils.print_info(data)
        typer.secho("\nAdvanced Configuration:", fg="blue")
        console.print(json.loads(mv.fit_config))
        typer.secho("\nFeature Importance:", fg="blue")
        console.print(
            {
                k: v
                for k, v in sorted(
                    mv.experiment.feature_importance.items(),
                    key=lambda item: item[1],
                    reverse=True,
                )
            }
        )
        typer.secho("\nData Checks:", fg="blue")
        console.print(mv.data_checks.to_dict().get("results"))


@app.command("promote")
@utils.exit_on_error
def promote(
    model_version: str = typer.Argument(..., help="Model version ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    model: str = typer.Option(None, help="Model ID."),
    wait: bool = typer.Option(False, "--wait", help="Wait for promotion to finish."),
    debug: bool = typer.Option(
        False, "--debug", help="Print full name of model version."
    ),
):
    """Promotes a model version."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.model_versions.get(model_version)
    d = mv.promote()
    if wait:
        d.wait()
        typer.secho(
            "Successfully promoted %s for model version %s." % (d.id, mv.id),
            fg="green",
        )
    else:
        typer.secho(
            "Started promotion %s for model version %s." % (d.id, mv.id),
            fg="green",
        )
    endpoint = utils.get_endpoint()
    model = mv.parent.split("/")[-1]
    typer.secho("You can review the promotion at: ", fg="blue")
    typer.secho(
        f"  {endpoint}/projects/{project}/model/{model}/promotions/\n", fg="blue"
    )

    if debug:
        typer.echo("DEBUG: Full name = %s." % d.name)


@app.command("cancel")
@utils.exit_on_error
def cancel(
    model_version: str = typer.Argument(..., help="Model version ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Cancel a model version that is currently training."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if not force:
        force = typer.confirm(
            "Are you sure you want to cancel training model version %s in project %s, environment %s?"
            % (model_version, project, environment)
        )
    if force:
        mv = c.model_versions.get(model_version)
        mv.cancel()
        typer.secho(
            "Successfully cancelled training for model version %s." % (mv.id),
            fg="green",
        )
    else:
        typer.secho("Skipping cancellation!", fg="blue")


@app.command("export")
@utils.exit_on_error
def export_yaml(
    model_version: str = typer.Argument(..., help="Model version ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    path: str = typer.Option(
        "print",
        help="Path to save YAML (local). Default = print to console.",
        show_default=False,
    ),
):
    """Export YAML file for model version."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.model_versions.get(model_version)
    m_text = yaml.load(mv.model_schema, Loader=yaml.FullLoader)
    yaml_text = utils.format_yaml(m_text)
    if path == "print":
        typer.secho(
            "Successfully extracted YAML file from model version %s." % model_version,
            fg="green",
        )
        print(yaml_text)
    else:
        utils.save_yamls([(model_version, yaml_text)], path)
        typer.secho(
            "Successfully extracted YAML file from model_version %s. Location is %s."
            % (model_version, path),
            fg="green",
        )


@app.command("compare-across-model")
@utils.exit_on_error
def compare_across_model(
    model: str = typer.Argument(..., help="Model ID.", show_default=False),
    project: str = typer.Option(None, help="Project ID.", show_default=False),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    model_versions: str = typer.Option(
        None, help="Model Version IDs, comma separated list", show_default=False
    ),
    performance_metric: utils.PerformanceMetric = typer.Option(
        None, help="Performance Metric to use for comparison."
    ),
    split_type: utils.SplitType = typer.Option(
        utils.SplitType.TEST, help="Dataset to use for comparison."
    ),
    stacked: bool = typer.Option(
        True, "--report", help="Print information as a report."
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """
    Compare model versions in a model.
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    try:
        m = c.models.get(model)
        if model_versions is not None:
            mvs = model_versions.split(",")
            mv_list = []
            for mv in mvs:
                model_version = m.model_versions.get(mv)
                mv_list.append(model_version)
        else:
            mv_list = c.model_versions.list_all()

        data = []
        metrics = {}
        fi = {}

        typer.secho(
            "Comparing model versions in model %s in project %s:" % (model, project)
        )
        for mv in mv_list:
            try:
                get_model_info(
                    project,
                    environment,
                    mv.id,
                    performance_metric,
                    split_type,
                    data,
                    metrics,
                    fi,
                )
            except:
                continue
        if len(data) > 1:
            print_mv_dashboard(
                data,
                metrics,
                fi,
                split_type,
                utils.get_style(style),
                include_project=False,
                stacked=stacked,
            )
        else:
            typer.secho(
                "Failed to find two or more successful model versions for model %s in project %s, environment %s."
                % (model, project, environment),
                fg="red",
            )

    except Exception as e:
        typer.secho(
            "Failed to compare model versions in model %s: %s" % (model, str(e)),
            fg="red",
        )


@app.command("compare")
@utils.exit_on_error
def compare(
    right_model_version: str = typer.Option(
        ...,
        "--right_model_version",
        "-rmv",
        help="Model Version ID.",
        show_default=False,
    ),
    right_project: str = typer.Option(
        None, "--right_project", "-rproj", help="Project ID.", show_default=False
    ),
    right_environment: str = typer.Option(
        None, "--right_environment", "-renv", help="Environment ID."
    ),
    left_model_version: str = typer.Option(
        ...,
        "--left_model_version",
        "-lmv",
        help="Model Version ID.",
        show_default=False,
    ),
    left_project: str = typer.Option(
        None, "--left_project", "-lproj", help="Project ID.", show_default=False
    ),
    left_environment: str = typer.Option(
        None, "--left_environment", "-lenv", help="Environment ID."
    ),
    performance_metric: utils.PerformanceMetric = typer.Option(
        None, help="Performance Metric to use for comparison."
    ),
    split_type: utils.SplitType = typer.Option(
        utils.SplitType.TEST, help="Data split to use for comparison."
    ),
    stacked: bool = typer.Option(
        True, "--report", help="Print information as a report."
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """Compare two model versions across projects or environments."""
    data = []
    metrics = {}
    fi = {}
    get_model_info(
        right_project,
        right_environment,
        right_model_version,
        performance_metric,
        split_type,
        data,
        metrics,
        fi,
        include_project=True,
    )
    get_model_info(
        left_project,
        left_environment,
        left_model_version,
        performance_metric,
        split_type,
        data,
        metrics,
        fi,
        include_project=True,
    )

    print_mv_dashboard(
        data,
        metrics,
        fi,
        split_type,
        utils.get_style(style),
        include_project=True,
        stacked=stacked,
    )


def print_mv_dashboard(
    data,
    metrics,
    fi,
    split_type,
    style,
    include_project=False,
    include_model=False,
    stacked=True,
):
    if include_project and include_model:
        df = pd.DataFrame(
            data,
            columns=[
                "Project ID",
                "MID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
            ],
        ).sort_values(["Perf Metric Value"], ascending=False)
    elif include_project:
        df = pd.DataFrame(
            data,
            columns=[
                "Project ID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
            ],
        ).sort_values(["Perf Metric Value"], ascending=False)
    elif include_model:
        df = pd.DataFrame(
            data,
            columns=[
                "MID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
            ],
        ).sort_values(["Perf Metric Value"], ascending=False)
    else:
        df = pd.DataFrame(
            data,
            columns=["MVID", "Created", "Algo", "Perf Metric", "Perf Metric Value"],
        ).sort_values(["Perf Metric Value"], ascending=False)

    df["Lead Perf"] = df["Perf Metric Value"].shift(-1)
    df["% Perf Change"] = (
        (df["Perf Metric Value"] - df["Lead Perf"]) / df["Perf Metric Value"] * 100
    )

    df["Lead MVID"] = df["MVID"].shift(-1)
    df["Features Dropped"] = df.apply(
        lambda x: "\n".join(
            sorted(
                list(
                    set(fi.get(x["Lead MVID"], fi.get(x["MVID"])))
                    - set(fi.get(x["MVID"]))
                )
            )
        ),
        axis=1,
    )
    df["Features Added"] = df.apply(
        lambda x: "\n".join(
            sorted(
                list(
                    set(fi.get(x["MVID"]))
                    - set(fi.get(x["Lead MVID"], fi.get(x["MVID"])))
                )
            )
        ),
        axis=1,
    )
    df["Top Features"] = df.apply(
        lambda x: "\n".join(
            list(
                "%s\t%s" % (y[0], y[1])
                for y in sorted(
                    fi.get(x["MVID"]).items(), key=lambda item: item[1], reverse=True
                )[:5]
            )
        ),
        axis=1,
    )

    metrics_df = pd.DataFrame(metrics).T.reset_index().rename(columns={"index": "MVID"})
    if include_project and include_model:
        main_table = utils.print_table(
            df[
                [
                    "Project ID",
                    "MID",
                    "MVID",
                    "Created",
                    "Algo",
                    "Perf Metric",
                    "Perf Metric Value",
                    "% Perf Change",
                ]
            ],
            [
                "Project ID",
                "MID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
                "% Perf Change",
            ],
            style=style,
            print=stacked,
            title="Model Version Comparison in %s data set" % split_type.value,
        )
    elif include_project:
        main_table = utils.print_table(
            df[
                [
                    "Project ID",
                    "MVID",
                    "Created",
                    "Algo",
                    "Perf Metric",
                    "Perf Metric Value",
                    "% Perf Change",
                ]
            ],
            [
                "Project ID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
                "% Perf Change",
            ],
            style=style,
            print=stacked,
            title="Model Version Comparison in %s data set" % split_type.value,
        )
    elif include_model:
        main_table = utils.print_table(
            df[
                [
                    "MID",
                    "MVID",
                    "Created",
                    "Algo",
                    "Perf Metric",
                    "Perf Metric Value",
                    "% Perf Change",
                ]
            ],
            [
                "MID",
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
                "% Perf Change",
            ],
            style=style,
            print=stacked,
            title="Model Version Comparison in %s data set" % split_type.value,
        )
    else:
        main_table = utils.print_table(
            df[
                [
                    "MVID",
                    "Created",
                    "Algo",
                    "Perf Metric",
                    "Perf Metric Value",
                    "% Perf Change",
                ]
            ],
            [
                "MVID",
                "Created",
                "Algo",
                "Perf Metric",
                "Perf Metric Value",
                "% Perf Change",
            ],
            style=style,
            print=stacked,
            title="Model Version Comparison in %s data set" % split_type.value,
        )
    metrics_table = utils.print_table(
        metrics_df,
        metrics_df.columns,
        style=style,
        print=stacked,
        title="Metrics Comparison",
    )
    top_features_table = utils.print_table(
        df[["MVID", "Top Features"]],
        ["MVID", "Top Features"],
        style=style,
        print=stacked,
        title="Feature Importance Comparison",
    )
    features_diff_table = utils.print_table(
        df[["MVID", "Features Dropped", "Features Added"]],
        ["MVID", "Features Dropped", "Features Added"],
        style=style,
        print=stacked,
        title="Feature Comparison",
    )

    if not stacked:
        features_diff_table = utils.print_table(
            df[["MVID", "Features Dropped", "Features Added"]],
            ["MVID", "Features Dropped", "Features Added"],
            style=style,
            print=False,
            title="Feature Comparison",
        )
        top_features_table = utils.print_table(
            df[["MVID", "Top Features"]],
            ["MVID", "Top Features"],
            style=style,
            print=False,
            title="Feature Importance Comparison",
        )
        metrics_table = utils.print_table(
            metrics_df,
            metrics_df.columns,
            style=style,
            print=False,
            title="Metrics Comparison",
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {"name": "Middle", "ratio": len(metrics_df), "data": metrics_table},
            {
                "name": "Bottom",
                "ratio": max(len(df), 5),
                "direction": "horizontal",
                "children": [
                    {"name": "Feature Comparison", "data": features_diff_table},
                    {"name": "Top Features", "data": top_features_table},
                ],
            },
        ]

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)


def get_model_info(
    project,
    environment,
    model_version,
    performance_metric,
    split_type,
    data=[],
    metrics={},
    fi={},
    include_project=False,
    include_model=False,
):
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.model_versions.get(model_version)

    if mv.state.value == "SUCCEEDED":  # filter out anything that didn't complete
        ex = mv.experiment

        perf_metric = "N/A"
        perf_metric_val = "N/A"
        if performance_metric is not None:
            perf_metric = performance_metric.value
            perf_metric_val = (
                ex.metrics.__getattribute__("%s_metrics" % (split_type.value))
                .__getattribute__(ex.model_type.value.lower())
                .__getattribute__(perf_metric)
            )
        else:
            perf_metric = mv.performance_metric
            perf_metric_val = mv.performance_metric_val
        arr = [
            mv.id,
            mv.update_time.strftime("%Y-%m-%d %H:%M:%S"),
            # mv.model_version_format,
            ex.type.split("/")[1],
            perf_metric,
            perf_metric_val,
        ]
        if include_model:
            arr.insert(0, mv.parent.split("/")[-1])
        if include_project:
            arr.insert(0, utils.get_environment_name(project, environment))
        data.append(arr)
        metrics[mv.id] = (
            ex.metrics.__getattribute__("%s_metrics" % (split_type.value))
            .__getattribute__(ex.model_type.value.lower())
            .to_dict()
        )
        fi[mv.id] = {
            ":".join(i.split(":")[1:]): j for i, j in ex.feature_importance.items()
        }

    return (data, metrics, fi)



import argparse
import logging
from glob import glob

from continual.python.sdk.client import Client
from continual.python.sdk.logger.logger import logger as metadata_logger
from continual.services.compute_services.pipelines.batchprediction.batch_prediction_runner import (
    BatchPredictionRunner,
)
from continual.services.compute_services.utils.compute_logger import (
    configure_logging,
    StandardOutputSplicer,
    StandardErrorSplicer,
)
from continual.python.utils.utils import get_parent


def main(args):
    client = Client(
        api_key=args.session,
        endpoint=args.management_endpoint,
        logging_resource=args.batch_prediction_name,
        project=get_parent(get_parent(args.batch_prediction_name)),
    )
    metadata_logger.init(
        resource=args.batch_prediction_name,
        session=args.session,
        endpoint=args.management_endpoint,
    )
    try:
        BatchPredictionRunner(
            client=client,
            batch_prediction_name=args.batch_prediction_name,
            logging_level=args.logging_level,
        ).run()
    finally:
        for log_file in glob("/app/*.log.txt"):
            client.logger.log_artifact(
                path=log_file, type="log", metadata={"name": log_file}
            )


if __name__ == "__main__":
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--logging-level",
            type=str,
            required=False,
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "FATAL"],
            help="Logging levels (equivalent to python logging module levels)",
        )
        parser.add_argument(
            "--batch-prediction-name",
            type=str,
            required=True,
            default="",
            help="The batch prediction to run.",
        )
        parser.add_argument(
            "--session",
            type=str,
            required=True,
            default="",
            help="The authenticared session to use for API calls",
        )
        parser.add_argument(
            "--management-endpoint",
            type=str,
            required=False,
            default="https://api.continual.ai",
            help="Address of the Management service.",
        )
        parser.add_argument(
            "--service-log-path",
            type=str,
            required=False,
            default="batchprediction.master.log.txt",
            help="Path to the file where all records from training service are logged.",
        )
        in_args = parser.parse_args()
        configure_logging(
            master_log_path=in_args.service_log_path,
        )
        with StandardOutputSplicer(
            log_path=in_args.service_log_path
        ), StandardErrorSplicer(log_path=in_args.service_log_path):
            main(in_args)
    except Exception as e:
        logging.error("Exception in main program: [{}]".format(e))
        raise e



import sqlalchemy as db
from sqlalchemy.exc import IntegrityError, ProgrammingError

import logging
import traceback
from json import loads, dumps
from google.oauth2 import service_account
from jinja2 import Template
from datetime import datetime
import pandas as pd
import os

from sqlalchemy.sql.schema import MetaData, Table
from continual.python.utils.utils import create_limit_query, is_select_query
from continual.rpc.management.v1 import management_types_pb2
from .prediction_store import PredictionStore
from .featurestore import FeatureStore, Typemap, FilterTables
from continual.python.sdk.exceptions import InvalidArgumentError

# from .query_templates import stats_template_bigquery
from continual.services.dataingest.gcs_source import FeatureSourceCloudStore
import uuid
from google.cloud import bigquery
from google.cloud.exceptions import NotFound

stats_template_bigquery = Template(
    """
 {% for elem in params.AllFeatures %} {% if loop.index > 1 %} UNION ALL {% endif %}
	SELECT {% if elem.Alias %} '{{elem.Alias}}' {% else %} '{{elem.name}}' {% endif %}as `column_name`,
{% if (elem.type == "NUMBER" or elem.type == "REAL" or elem.type == "FIXED") %}
	COALESCE(max(allentities.{{elem.name}}), 0.0) as `max`, COALESCE(min(allentities.{{elem.name}}), 0.0) as `min`,
	COALESCE(avg(allentities.{{elem.name}}), 0.0) as `mean`, COALESCE(stddev(allentities.{{elem.name}}), 0.0) as `sd`,
	0 as `min_length`, 0 as `max_length`, 0 as `distinct_values`, CAST(NULL as timestamp)  as `latest`, CAST(NULL as timestamp)  as `earliest`,
{% elif (elem.type == "TIMESTAMP" or elem.type == "TIMESTAMP_LTZ" or elem.type == "TIMESTAMP_TZ" or elem.type == "TIMESTAMP_NTZ" or elem.type == "DATE") %}
	0.0 as `max`, 0.0 as `min`, 0.0 as `mean`, 0.0 as `sd`, 0 as `min_length`, 0 as `max_length`, 0 as `distinct_values`, 
    cast(max(allentities.{{elem.name}}) as timestamp) as `latest`, cast(min(allentities.{{elem.name}}) as timestamp) as `earliest`,
{% elif (elem.type == "TEXT" or elem.type == "STRING") %}
	0 as `max`, 0 as `min`, 0.0 as `mean`, 0.0 as `sd`, COALESCE(min(length(allentities.{{elem.name}})), 0) as `min_length`,
	COALESCE(max(length(allentities.{{elem.name}})), 0) as `max_length`, COALESCE(count(distinct(allentities.{{elem.name}})), 0) as `distinct_values`,
	CAST(NULL as timestamp)  as `latest`, CAST(NULL as timestamp)  as `earliest`,
{% elif elem.type == "CATEGORICAL" %}
	0 as `max`, 0 as `min`, 0.0 as `mean`, 0.0 as `sd`,
	COALESCE(min(length(allentities.{{elem.name}})), 0) as `min_length`, COALESCE(max(length(allentities.{{elem.name}})), 0) as `max_length`,
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as `distinct_values`, CAST(NULL as timestamp)  as `latest`,
	CAST(NULL as timestamp)  as `earliest`,
{% elif elem.type == "BOOLEAN" %}
	0 as `max`, 0 as `min`, 0.0 as `mean`, 0.0 as `sd`, 0 as `min_length`, 0 as `max_length`,
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as `distinct_values`, CAST(NULL as timestamp)  as `latest`, CAST(NULL as timestamp)  as `earliest`,
{% else %}
	0 as `max`, 0 as `min`, 0.0 as `mean`, 0.0 as `sd`, 0 as `min_length`, 0 as `max_length`, 0 as `distinct_values`,
	CAST(NULL as timestamp)  as `latest`, CAST(NULL as timestamp)  as `earliest`,{% endif %}
	count(*) - count(allentities.{{elem.name}})  as `null_count`,
	count(allentities.{{elem.name}}) as `count`,
	CURRENT_TIMESTAMP as `computed_at`, '{{elem.ctype}}' as `type`
    FROM  {{params.FullName}} as allentities
	{% endfor %}
    """
)


class BigQueryClient:
    def __init__(self, data_store):
        self.data_store = data_store
        if data_store.big_query.auth_file is None:
            return InvalidArgumentError("Invalid GCS credentials file")

        bq_cred = loads(data_store.big_query.auth_file)
        if "project_id" not in bq_cred:
            raise InvalidArgumentError("Invalid GCS credentials file")

        self.client_uuid = str(uuid.uuid4().hex)
        self.project = bq_cred["project_id"]
        logging.info("set bq project to: {}".format(self.project))
        self.cred_file = f"/tmp/{bq_cred['project_id']}_{self.client_uuid}.json"
        with open(self.cred_file, "w") as cred_file:
            cred_file.write(data_store.big_query.auth_file)

        self.credentials = service_account.Credentials.from_service_account_file(
            self.cred_file,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

    def __enter__(self):
        try:
            self.client = bigquery.Client(
                project=self.project, credentials=self.credentials
            )
            return self.client, self.project, self.cred_file
        except:
            logging.info(traceback.format_exc())
            raise

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self.client.close()
        if os.path.isfile(self.cred_file):
            os.remove(self.cred_file)


def create_connection(data_store, dataset=None):
    """
    Create bigquery connection given data_store.
    """
    with BigQueryClient(data_store) as (_, project, cred_file):
        logging.info("using bg project for connection: {}".format(project))
        eng = db.create_engine(
            "bigquery://{project}".format(project=project), credentials_path=cred_file
        )

        connection = eng.connect()
        return eng, connection, project


class BigqueryFeatureStore(FeatureStore, PredictionStore):

    """
    BigqueryFeatureStore implements both source of features and featurestore functions.
    Currently sources are not pooled. In a larger tenant with frequent and multiple fetches,
    we would need to consider connection pooling.
    """

    DTYPE_TO_LOGICAL = {
        "FIXED": "NUMBER",
        "FLOAT": "NUMBER",
        "FLOAT64": "NUMBER",
        "INT": "NUMBER",
        "INT64": "NUMBER",
        "SMALLINT": "NUMBER",
        "SMALLINT": "NUMBER",
        "TINYINT": "NUMBER",
        "BYTEINT": "NUMBER",
        "INTEGER": "NUMBER",
        "NUMERIC": "NUMBER",
        "DECIMAL": "NUMBER",
        "BIGNUMERIC": "NUMBER",
        "BIGDECIMAL": "NUMBER",
        "TEXT": "CATEGORICAL",
        "STRING": "CATEGORICAL",
        "DATE": "TIMESTAMP",
        "DATETIME": "TIMESTAMP",
        "TIMESTAMP": "TIMESTAMP",
        "VARIANT": "TEXT",
        "TIMESTAMP_LTZ": "TIMESTAMP",
        "TIMESTAMP_TZ": "TIMESTAMP",
        "TIMESTAMP_NTZ": "TIMESTAMP",
        "OBJECT": "TEXT",
        "ARRAY": "STRING_ARRAY",
        "BYTES": "TEXT",
        "TIME": "TIMESTAMP",
        "BOOLEAN": "BOOLEAN",
        "BOOL": "BOOLEAN",
    }

    def __init__(self, data_store, source=None, cfg=None):
        self.data_store = data_store
        self.source = source
        self.corpus_id = None
        self.client = None

        try:
            self.engine, self.connection, self.bq_project = create_connection(
                data_store
            )
        except Exception:
            raise InvalidArgumentError(
                "Connection to bigquery failed, please check connection settings.",
                details=None,
            )
        super(BigqueryFeatureStore, self).__init__()

    @property
    def location(self) -> str:
        with BigQueryClient(self.data_store) as (client, _, _):
            default_location = self.data_store.big_query.location or "US"
            dataset_id = "{}.{}".format(client.project, self.db_schema)
            try:
                dataset = client.get_dataset(dataset_id)
                return dataset.location or default_location
            except NotFound:
                return default_location

    def cleanup(self):
        with BigQueryClient(self.data_store) as (client, _, _):
            dataset_id = "{}.{}".format(client.project, self.db_schema)
            client.delete_dataset(
                dataset_id, delete_contents=True, not_found_ok=True
            )  # Make an API request.
            return

    def close(self):
        self.connection.close()

    def get_type(self) -> str:
        return "bigquery"

    def is_hosted(self):
        return False

    def get_stats_query_template(self):
        return stats_template_bigquery

    def create_prediction_model_table(
        self, model, problem_type, prev_index, prev_time_index
    ):
        """
        Create or update the table used to store predictions.

        This checks to see if the prediction table for this model exists, if it doesn't it creates it. If it does, it updates it if necessary.

        Parameters
        ----------
        model : continual.rpc.management.v1.Model
            The Model definition.
        problem_type : string
            The type of problem the model is solving.  E.g. regression, multiclass_classification, binary_classification.
        prev_index: string
            If there already is a model schema defined, the value of the previous model schema definition's index. Empty string if there is no previous model definition.
        prev_time_index: string
            If there already is a model schema defined, the value of the previous model schema definition's time index. Empty string if there is no previous model definition.
        """
        meta = db.MetaData()

        meta.reflect(bind=self.engine, only=FilterTables(schema=self.db_schema))
        _, model_pred_name = self.get_model_prediction_table_name(model)
        model_table = None
        for t in meta.sorted_tables:
            if f"{self.db_schema}.{model_pred_name}" == t.name:
                model_table = t
        if model_table is None:
            self._create_prediction_model_table(
                model, model_pred_name, meta, problem_type
            )
        else:
            self._update_prediction_model_table(
                model, model_table, problem_type, prev_index, prev_time_index
            )

    def _create_prediction_model_table(
        self, model, model_pred_name, meta: MetaData, problem_type
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.
        Args: Feature set for which the tables are to be created.≈
        Returns None
        """
        index = None
        time_index = None
        target_column = model.schema.target
        target_column_dtype = db.TEXT
        for f in model.schema.columns:
            if (
                f.type == management_types_pb2.FieldType.INDEX
                or f.name == model.schema.index
            ):
                index = f
            elif (
                f.type == management_types_pb2.FieldType.TIME_INDEX
                or f.name == model.schema.time_index
            ):
                time_index = f
            elif f.name == target_column:
                target_column_dtype = Typemap[f.dtype]
                # a regression prediction is always a floating point number
                if problem_type == "regression":
                    target_column_dtype = db.FLOAT

        columns = []
        columns.append(db.Column("batch_prediction", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP(timezone=True)))
        columns.append(db.Column(f"features", db.Text))

        if index is not None:
            columns.append(db.Column(index.name, Typemap[index.dtype]))

        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.dtype]))
        columns.append(db.Column(f"{target_column}_prediction", target_column_dtype))
        if problem_type == "multiclass_classification":
            columns.append(db.Column(f"{target_column}_prediction_details", db.Text))
            columns.append(db.Column(f"{target_column}_prediction_score", db.Float))
        if problem_type == "binary_classification":
            columns.append(
                db.Column(f"{target_column}_true_prediction_score", db.Float)
            )

        db.Table(model_pred_name, meta, *columns, schema=self.db_schema)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    raise

    def _update_prediction_model_table(
        self, model, model_table: Table, problem_type, prev_index, prev_time_index
    ):

        logging.info(f"update table: {model_table.name}")
        index_column = model.schema.index
        time_index_column = model.schema.time_index
        target_column = model.schema.target
        prediction_col = f"{target_column}_prediction"
        prediction_details_col = f"{target_column}_prediction_details"
        prediction_score_col = f"{target_column}_prediction_score"
        if problem_type == "binary_classification":
            prediction_score_col = f"{target_column}_true_prediction_score"

        if (
            prediction_col in model_table.columns.keys()
            and (index_column == "" or index_column in model_table.columns.keys())
            and (
                time_index_column == ""
                or time_index_column in model_table.columns.keys()
            )
        ):
            # no changes to make
            logging.info("no changes to make to table")
            return
        # rename columns
        old_prediction_col = ""
        old_prediction_details_col = ""
        old_prediction_score_col = ""
        for col_name in model_table.columns.keys():
            if col_name.endswith("_prediction"):
                old_prediction_col = col_name
            if col_name.endswith("_prediction_details"):
                old_prediction_details_col = col_name
            if col_name.endswith("_prediction_score"):
                old_prediction_score_col = col_name

        if old_prediction_col == "":
            raise ValueError(
                f"predictions columms not found in table {model_table.name}"
            )

        indexes_except = ""
        indexes_as = ""

        if index_column != "" and index_column != prev_index:
            indexes_except = f", {prev_index}"
            indexes_as = f", {prev_index} as {index_column}"
        if time_index_column != "" and time_index_column != prev_time_index:
            indexes_except = f"{indexes_except}, {prev_index}"
            indexes_as = f"{indexes_as}, {prev_index} as {index_column}"

        # TODO: fix so we do time index and or index as appropriate
        with BigQueryClient(self.data_store) as (bqc, _, _):
            if problem_type == "multiclass_classification":
                alter_query = f""" CREATE OR REPLACE TABLE `{model_table.name}`  AS
                        SELECT * EXCEPT({old_prediction_col}, {old_prediction_details_col}, {old_prediction_score_col} {indexes_except}) {indexes_as}, {old_prediction_col} as {prediction_col}, 
                        {old_prediction_details_col} as {prediction_details_col}, {old_prediction_score_col} as {prediction_score_col} 
                        FROM `{model_table.name}`"""
                query_job = bqc.query(alter_query)
                query_job.result()
            if problem_type == "binary_classification":
                alter_query = f""" CREATE OR REPLACE TABLE `{model_table.name}`  AS
                        SELECT * EXCEPT({old_prediction_col}, {old_prediction_score_col} {indexes_except}) {indexes_as}, {old_prediction_col} as {prediction_col}, 
                        {old_prediction_score_col} as {prediction_score_col} 
                        FROM `{model_table.name}`"""
                query_job = bqc.query(alter_query)
                query_job.result()
            if problem_type == "regression":
                alter_query = f""" CREATE OR REPLACE TABLE `{model_table.name}`  AS
                        SELECT * EXCEPT({old_prediction_col} {indexes_except}) {indexes_as}, {old_prediction_col} as {prediction_col}
                        FROM `{model_table.name}`"""

                query_job = bqc.query(alter_query)
                query_job.result()

    def create_project_schema(self):
        if self.db_schema == None:
            raise InvalidArgumentError("schema_name is required")

        with BigQueryClient(self.data_store) as (client, _, _):
            dataset_id = "{}.{}".format(client.project, self.db_schema)

            try:
                client.get_dataset(dataset_id)  # Make an API request.
            except NotFound:
                logging.debug("Dataset {} is not found".format(dataset_id))
                # Construct a full Dataset object to send to the API.
                dataset = bigquery.Dataset(dataset_id)
                dataset.location = self.location

                # Send the dataset to the API for creation, with an explicit timeout.
                # Raises google.api_core.exceptions.Conflict if the Dataset already
                # exists within the project.
                dataset = client.create_dataset(
                    dataset, timeout=30
                )  # Make an API request.
        return

    def get_database_names(self):
        with BigQueryClient(self.data_store) as (client, _, _):
            projects = []
            for d in client.list_projects():
                projects.append(d.project_id)
            return projects

    def get_schema_names(self, database):
        with BigQueryClient(self.data_store) as (client, _, _):
            schemas = []
            for d in client.list_datasets(project=client.project):
                schemas.append(d.dataset_id)

            return schemas

    def get_table_names(self, database, schema):
        with BigQueryClient(self.data_store) as (client, _, _):
            tables = []
            for d in client.list_tables(dataset=schema):
                tables.append(d.table_id)

            return tables

    def get_column_names(self, database, schema, table):
        columns = []
        type_map = {}

        with BigQueryClient(self.data_store) as (client, _, _):
            dataset = client.dataset(schema)
            table_ref = dataset.table(table)

            bq_table = client.get_table(table_ref)
            logging.debug(f"table properties: {bq_table._properties}")

            fields = bq_table._properties["schema"]["fields"]
            for field in fields:
                name, field_type = field["name"], field["type"]
                columns.append(name)
                type_map[name] = field_type

        return columns, type_map

    def load_data(self, filename, table_name, replace=False):
        with BigQueryClient(self.data_store) as (_, bqproject, key_file_name):
            if bqproject is None:
                return bigquery.Client()

            credentials = service_account.Credentials.from_service_account_file(
                key_file_name,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )

            source = FeatureSourceCloudStore(str(uuid.uuid4()), filename)
            df = source.fetch_all()
            exists = "fail"
            if replace:
                exists = "replace"
            name = "{}.{}".format(self.db_schema, table_name)
            try:
                df.to_gbq(
                    name,
                    project_id=bqproject,
                    chunksize=10000,
                    if_exists=exists,
                    credentials=credentials,
                )
            except ValueError:
                traceback.print_exc()
                raise

            return table_name

    def infer_schema(
        self,
        query,
        excludes=None,
        sample_row_count=100,
        allow_invalid_types=False,
        timeout_ms=0,
    ):
        inferred_schema = {}
        all_rows = []
        with BigQueryClient(self.data_store) as (client, _, _):
            if is_select_query(query):
                query = create_limit_query(query, sample_row_count or 100)
            logging.info("About to run query: {}".format(query))

            query_job = client.query(query, location=self.location)
            res = query_job.result()

            columns = query_job._query_results._properties["schema"]["fields"]
            for col in columns:
                col_name = col["name"].lower()
                if excludes and col_name in excludes:
                    continue
                dtype = col["type"]
                if allow_invalid_types and dtype not in self.DTYPE_TO_LOGICAL:
                    inferred_schema[col_name] = {
                        "type": "INVALID",
                        "dtype": dtype,
                    }
                elif dtype in self.DTYPE_TO_LOGICAL:
                    inferred_schema[col_name] = {
                        "type": self.DTYPE_TO_LOGICAL[dtype],
                        "dtype": dtype,
                    }

            for row in res:
                drows = dict(row)
                # Filter our columns in the exclude list.
                if excludes is not None:
                    for e in excludes:
                        if e in drows:
                            del drows[e]

                all_rows.append(dumps(drows, default=str))
                if len(all_rows) >= sample_row_count:
                    break

        return inferred_schema, all_rows

    def _process_dataframe(
        self, df, id_column_name, ts_column_name, name, model_version
    ):
        for col in df.columns:
            if col.startswith("Unnamed:"):
                df = df.drop(col, axis=1)

        df.columns = map(str.lower, df.columns)
        # new query object should support better implementation
        for col in df.columns:
            if col.endswith(f":{id_column_name}"):
                df = df.rename(columns={col: id_column_name})
            if col.endswith(f":{ts_column_name}"):
                df = df.rename(columns={col: ts_column_name})

        df["batch_prediction"] = name
        df["model_version"] = model_version

        df = df.where(pd.notnull(df), None)

        return df

    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.debug("in write_prediction")
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )

        df["prediction_time"] = job_start_time
        full_name = f"{self.db_schema}.{pred_table}"

        with BigQueryClient(self.data_store) as (client, bqproject, key_file_name):
            credentials = service_account.Credentials.from_service_account_file(
                key_file_name,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )

            ## Insert into bigquery
            logging.info("df columns: {}".format(df.columns))
            logging.info("df columns dtypes: {}".format(df.dtypes))
            logging.info("full_name: {}".format(full_name))
            logging.info("time_index_dtype: {}".format(time_index_dtype))

            table_schema = []
            if time_index_dtype != None and (
                time_index_dtype == "DATE"
                or time_index_dtype == "DATETIME"
                or time_index_dtype == "TIME"
            ):
                table_schema.append({"name": ts_column_name, "type": time_index_dtype})
                logging.info(
                    "updated ts_column_name type to: {}".format(time_index_dtype)
                )

            df.to_gbq(
                full_name,
                bqproject,
                chunksize=None,
                if_exists="append",
                credentials=credentials,
                table_schema=table_schema,
            )

            count_query = f"SELECT COUNT(1) from `{self.db_schema}.{pred_table}` WHERE batch_prediction = '{job_name}'"
            res = client.query(count_query, location=self.location)
            for row in res:
                return row[0]
            return 0



import typer
import os
from pathlib import Path

from continual.python.cli import utils
from continual.python.sdk.client import Client


app = typer.Typer(help="Manage CLI configuration.")

# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        show()


@app.command("init")
@utils.exit_on_error
def continual_init(
    project: str = typer.Option(..., help="Project ID."),
    environment: str = typer.Option(None, help="Environment Id"),
    style: utils.ContinualStyle = typer.Option(None, help="Display style."),
):
    """Set default project.

    Sets the default project for the SDK and CLI.  All future commands
    will use the default project if no project is explicitly passed. You may
    set the project and environment at the same time via project@environment.
    """
    try:
        config = Client().config
        Path(os.path.join(os.getcwd(), ".continual.yaml")).touch(exist_ok=True)
        c = Client()
        c.config.email = config.email
        c.config.endpoint = config.endpoint
        c.config.api_key = config.api_key
        c.config.save()
        set_project(project=project, print=False)
        if environment:
            set_environment(environment=environment, print=False)
        else:
            c.config.environment = config.environment
        if style:
            set_style(style=style, print=False)
        else:
            c.config.style = config.style
        c = Client()
        c.config.show()
    except:
        typer.secho(
            "Project %s does not exist. Please create the project first!" % project,
            fg="red",
        )


@app.command("set-project")
@utils.exit_on_error
def set_project(
    project: str = typer.Argument(..., help="Project ID."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default project.

    Sets the default project for the SDK and CLI.  All future commands
    will use the default project if no project is explicitly passed. You may
    set the project and environment at the same time via project@environment.
    """
    try:
        c = Client(project=project)
        if project == "":
            project = None
        if project is not None and "/" not in project:
            project = "projects/" + project
        if "@" in project:
            environment = project.split("@")[1]
            c.config.environment = environment
            project = project.split("@")[0]
        c.projects.get(project)
        c.config.project = project
        c.config.save()
        if print:
            c.config.show()
    except:
        typer.secho(
            "Project %s does not exist. Please create the project first!" % project,
            fg="red",
        )


@app.command("set-endpoint")
@utils.exit_on_error
def set_endpoint(
    endpoint: str = typer.Argument(..., help="Continual environment endpoint"),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default endpoint"""
    c = Client()
    old_endpoint = c.config.endpoint
    c.config.endpoint = endpoint
    c.config.save()

    # Reload client to check endpoint validity
    try:
        c = Client()
        typer.secho(f"Sucessfully set endpoint to {endpoint}", fg="green")
        if print:
            c.config.show()
    except ValueError as e:
        c.config.endpoint = old_endpoint
        c.config.save()
        typer.secho("Endpoint %s is not valid : %s" % (endpoint, e), fg="red")


@app.command("set-environment")
@utils.exit_on_error
def set_environment(
    environment: str = typer.Argument(..., help="Environment ID."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default environment.

    Sets the default environment for the SDK and CLI.  All future commands
    will use the default environment if no environment is explicitly passed.
    """
    try:
        c = Client()
        if environment == "":
            environment = None
        if environment and "@" in environment:
            environment = environment.split("@")[-1]
        c.environments.get(environment)
        c.config.environment = environment
        c.config.save()
        if print:
            c.config.show()
    except:
        typer.secho(
            "Environment %s does not exist. Please create the environment first!"
            % environment,
            fg="red",
        )


@app.command("set-style")
@utils.exit_on_error
def set_style(
    style: utils.ContinualStyle = typer.Argument(..., help="Display style."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default CLI display style."""
    c = Client()
    c.config.style = style.value
    c.config.save()
    if print:
        c.config.show()


@app.command("show")
@utils.exit_on_error
def show():
    """Show current config.

    Shows the current session configuration stored in
    ~/.continual/continual.yaml."
    """
    c = Client()
    c.config.show()



import typer
from continual.python.sdk.client import Client
from pprint import pprint
from continual.python.cli import utils

app = typer.Typer(help="Manage users.")

# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30)


@app.command("list")
@utils.exit_on_error
def list(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
):
    """List users."""
    print("ID - Name")
    c = Client()
    for user in c.users.list_all():
        typer.echo("%s - %s %s" % (user.id, user.first_name, user.last_name))


@app.command("get")
@utils.exit_on_error
def get(
    user_id: str = typer.Argument(..., help="User ID."),
    json: bool = typer.Option(False, "--json", help="Print full details."),
):
    """Get user details."""
    c = Client()
    user = c.users.get(user_id)
    if json:
        pprint(user.to_dict())
    else:
        typer.echo("ID: %s" % user.id)
        typer.echo("Name: %s %s" % (user.first_name, user.last_name))
        typer.echo("Email: %s" % user.email)
        typer.echo("Created: %s" % user.create_time)
        typer.echo("Updated: %s" % user.update_time)


@app.command("register")
@utils.exit_on_error
def register(
    first_name: str = typer.Argument(..., help="First name.", show_default=False),
    last_name: str = typer.Argument("", help="Last name.", show_default=False),
    email: str = typer.Option(
        ..., help="Email address.", prompt=True, show_default=False
    ),
    password: str = typer.Option(
        ..., help="Password.", prompt=True, hide_input=True, show_default=False
    ),
):
    """Create user."""
    c = Client()
    user = c.register(first_name, last_name, email, password)
    print("Created user %s %s with id %s." % (first_name, last_name, user.id))


@app.command("delete")
@utils.exit_on_error
def delete(
    user_id: str = typer.Argument(..., help="User ID."),
):
    """Delete user."""
    c = Client()
    try:
        user = c.users.get(user_id)
        user.delete()
        typer.echo("Successfully deleted user %s." % user_id)
    except Exception as e:
        typer.echo("Failed to deleted user %s: %s" % (user_id, str(e)))


@app.command("update")
@utils.exit_on_error
def update(
    user_id: str = typer.Argument(..., help="User ID."),
    first_name: str = typer.Option(..., help="First name."),
    last_name: str = typer.Option(..., help="Last name."),
):
    """Update Project Name"""
    c = Client()
    try:
        user = c.users.get(user_id)
        user.update(first_name, last_name)
        typer.echo("Successfully updated user %s." % user_id)
    except Exception as e:
        typer.echo("Failed to update user %s: %s" % (user_id, str(e)))



from continual.python.blob.gcs import GoogleCloudStorage


class BlobStore(object):

    SUPPORTED_STORES = [GoogleCloudStorage]

    def __init__(self, resource_url, **kwargs):
        self.resource_url = resource_url
        self._kwargs = kwargs
        self.store = self._validate_url(resource_url)

    def _validate_url(self, resource_url):
        ValidStore = None
        for s in self.SUPPORTED_STORES:
            if s.is_valid_url(resource_url):
                ValidStore = s
                break

        if not ValidStore:
            raise ValueError(
                "Invalid URL. No supported blob stores support this resource URL."
            )

        return ValidStore(resource_url, **self._kwargs)

    def list_objects(self):
        if not self.store:
            raise ValueError(
                "Cannot list objects - blob store has not been initialized."
            )

        return self.store.list_objects()

    def upload_file(self, filepath, **kwargs):
        if not self.store:
            raise ValueError("Cannot upload - blob store has not been initialized.")

        return self.store.upload_file(filepath, **kwargs)

    def upload_dir(self, dirpath, **kwargs):
        if not self.store:
            raise ValueError("Cannot upload - blob store has not been initialized.")

        return self.store.upload_dir(dirpath, **kwargs)

    def download_file(self, filepath=None, **kwargs):
        if not self.store:
            raise ValueError("Cannot download — blob store has not been initialized.")

        return self.store.download_file(filepath, **kwargs)

    def download_as_string(self, filepath=None, **kwargs):
        if not self.store:
            raise ValueError("Cannot download — blob store has not been initialized.")

        return self.store.download_as_string(filepath, **kwargs)

    def download_dir(self, dirpath, **kwargs):
        if not self.store:
            raise ValueError("Cannot download — blob store has not been initialized.")

        return self.store.download_dir(dirpath, **kwargs)



from __future__ import annotations
from typing import List, Optional, Iterator
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.waiter import wait
from continual.python.sdk.events import EventManager


class IngestionManager(Manager):
    """Manages ingestions resources."""

    name_pattern: str = (
        "projects/{project}/featureSets/{feature_set}/ingestions/{ingestions}"
    )

    def get(self, id: str) -> Ingestion:
        """Get ingestion.

        Arguments:
            id: Ingestion name or id.

        Returns
            A ingestion.
        """
        req = management_pb2.GetIngestionRequest(name=self.name(id))
        resp = self.client._management.GetIngestion(req)
        return Ingestion.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[Ingestion]:
        """List ingestions.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of ingestions.
        """
        req = management_pb2.ListIngestionsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListIngestions(req)
        return [Ingestion.from_proto(u, client=self.client) for u in resp.ingestions]

    def list_all(self) -> Iterator[Ingestion]:
        """List all ingestions.

        Pages through all ingestions using an iterator.

        Returns:
            A iterator of all ingestions.
        """

        def next_page(next_page_token):
            req = management_pb2.ListIngestionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListIngestions(req)
            return (
                [Ingestion.from_proto(u, client=self.client) for u in resp.ingestions],
                resp.next_page_token,
            )

        return Pager(next_page)

    def cancel(self, id: str) -> None:
        """Cancel an ingestion.

        Arugments:
            id: Name or id of ingestion

        """
        req = management_pb2.CancelIngestionRequest(name=self.name(id))
        self.client._management.CancelIngestion(req)


class Ingestion(Resource, types.Ingestion):
    """Ingestion resource."""

    name_pattern: str = (
        "projects/{project}/featureSets/{feature_set}/ingestions/{ingestions}"
    )
    manager: IngestionManager

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = IngestionManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def cancel(self):
        """Cancel ingestion."""
        self.manager.cancel(self.id)

    def wait(self, echo: bool = True, timeout: int = None) -> Ingestion:
        """Wait for ingestion to complete.

        Arguments:
            echo: Display progress.
            timeout: Timeout in seconds.
        Returns:
            Updated ingestion.
        """
        return wait(
            getter=lambda: self.manager.get(self.id),
            echo=echo,
            timeout=timeout,
        )



from continual.services.dataingest.server import DataIngestService
from continual.rpc.dataingest.v1 import dataingest_pb2_grpc
import grpc
import logging
from concurrent import futures
from http.server import HTTPServer, BaseHTTPRequestHandler
import threading
import time

CONFIG_ENV = "CONFIG_ENV"
CONFIG_LOCAL_BASE = "CONFIG_LOCAL_BASE"


def start_health_server(health_path="/healthz", shutdown_path="/shutdown", port=8000):
    logging.info(f"starting health server on port {port}")
    health_server = HealthHTTPRequestHandler(health_path, shutdown_path)
    httpd = HTTPServer(("", port), health_server)
    httpd.serve_forever()


class HealthHTTPRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, health_path, shutdown_path):
        self.health_path = health_path
        self.shutdown_path = shutdown_path
        self.alive = True

    def __call__(self, *args, **kwargs):
        """Handle a request."""
        super().__init__(*args, **kwargs)

    def do_GET(self):
        # currently this health check does not validate anything other than it has been called after
        # the grpc server has started, allowing kubernetes to wait until this has occured.
        # currently the grpc server does no startup tasks, if some are added, this can be modified
        # to wait and check for success
        logging.debug(
            f"alive: {self.alive} health_path: {self.health_path} shutdown_path: {self.shutdown_path} - process get: {self.path}"
        )
        if self.path == self.health_path and self.alive:
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b'"health": "ok"')
        elif self.path == self.shutdown_path:
            self.alive = False
            logging.info("wait 5 seconds before shutting down")
            time.sleep(5)
            logging.info("shutting down")
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b'"shutdown": "ok"')
        else:
            self.send_response(503)
            self.end_headers()
            self.wfile.write(b'"health": "invalid path"')


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    dataingest_pb2_grpc.add_DataIngestAPIServicer_to_server(DataIngestService(), server)
    server.add_insecure_port("[::]:3000")
    logging.info("server starting")
    server.start()

    # Start health server
    port = 5011
    health_daemon = threading.Thread(
        name="health_daemon_server",
        target=start_health_server,
        args=("/healthz", "/shutdown", port),
    )
    health_daemon.setDaemon(
        True
    )  # Set as a daemon so it will be killed once the main thread is dead.
    health_daemon.start()

    server.wait_for_termination()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    # for root, dirs, files in os.walk("/", topdown=False):
    #    for name in files:
    #        print(os.path.join(root, name))
    logging.info("starting service")
    serve()



import importlib

from continual.python.sdk.extensions.models.base_model_pipeline import BaseModelPipeline
from continual.python.sdk.client import Client

CONTINUAL_AUTOGLUON_PIPELINE_MODULE = "continual.python.engine.auto_classification_regression.continual_autogluon_pipeline"
CONTINUAL_AUTOGLUON_PIPELINE_CLASS = "ContinualAutogluonPipeline"
DEFAULT_PIPELINES = {
    "AutoClassificationRegression": {
        "module": CONTINUAL_AUTOGLUON_PIPELINE_MODULE,
        "class": CONTINUAL_AUTOGLUON_PIPELINE_CLASS,
    }
}


def get_pipeline(client: Client, pipeline_id: str = None) -> BaseModelPipeline:
    """
    Install load and instantiate the class for the custom pipeline
    """
    if not pipeline_id:
        pipeline_module = importlib.import_module(CONTINUAL_AUTOGLUON_PIPELINE_MODULE)
        pipeline_class = getattr(
            pipeline_module, CONTINUAL_AUTOGLUON_PIPELINE_CLASS, None
        )
        return pipeline_class

    elif pipeline_id in DEFAULT_PIPELINES:
        pipeline_module = importlib.import_module(
            DEFAULT_PIPELINES[pipeline_id]["module"]
        )
        pipeline_class = getattr(
            pipeline_module, DEFAULT_PIPELINES[pipeline_id]["class"], None
        )
        return pipeline_class

    else:
        # Use SDK to download extension
        client.extensions.install(id=pipeline_id)

        pipeline_extension = client.extensions.get(id=pipeline_id)

        # Import pipeline class using extension metadata
        pipeline_module = importlib.import_module(pipeline_extension.module_name)
        pipeline_class = getattr(pipeline_module, pipeline_extension.class_name, None)

        return pipeline_class



import logging


class Utilities:
    """
    Base class for utilities provided by the sdk such as a logger,
    artifact management etc.
    """

    def logger(self) -> logging.Logger:
        """
        Get a python logger instance
        to log in custom classes
        """
        raise NotImplementedError

    def save_artifact(self, source_path: str, description: str = None):
        """
        Save a custom artifact to source_path with optional description

        Parameters
        ----------
        source_path : str
            location of a file on disk
        description : str
            An optional descrption to associate with the saved artifact

        """
        raise NotImplementedError

    def save_pyplot_png(self, filename: str, display: bool):
        """
        Save the current figure in matplotlib as a png artifact
        Gets title and description from pyplot.

        Parameters
        ----------
        filename: str
            Name to save the png
        display: bool
            Whether to show the plot on the Continual UI
        """
        raise NotImplementedError

    def display_plot_from_file(self, title: str, source_path: str):
        """
        Display a plot image from disk on the Continual UI.

        Parameters
        ----------
        title: str
            Title for the plot
        source_path: str
            location of a file on disk
        """

        raise NotImplementedError



import traceback


from continual.python.sdk.client import Client
from continual.python.sdk.datasets import Dataset
from continual.python.sdk.extensions.models.base_model_pipeline import BaseModelPipeline
from continual.services.compute_services.pipelines.pipeline_utils import get_pipeline
from continual.services.compute_services.utils.compute_logger import (
    get_logger,
)

LOG_FILE_NAME = "batchprediction.runner.log.txt"


class BatchPredictionRunner:
    def __init__(
        self, client: Client, batch_prediction_name: str, logging_level: str = "INFO"
    ):

        self.client = client
        self.config = client.batch_prediction_jobs.get_batch_prediction_config(
            batch_prediction_name=batch_prediction_name
        )
        self.logging_level = logging_level

        # Extract config parameters
        self.batch_prediction_name = batch_prediction_name
        self.model_version_name = self.config.model_version
        self.model = self.config.model.to_proto()
        self.model_schema = self.model.schema
        self.data_store = self.config.data_store.to_proto()
        self.incremental = self.config.incremental
        self.model_version_signature = self.config.model_version_signature
        if (
            self.model_schema.train
            and self.model_schema.train.disable_feature_timestamp_generation
        ):
            self.disable_feature_timestamp_generation = (
                self.model_schema.train.disable_feature_timestamp_generation
            )
        else:
            self.disable_feature_timestamp_generation = False

        self.logger = get_logger(__name__, LOG_FILE_NAME, self.logging_level)

        self.featureset_map = {}
        for fs in self.config.feature_sets:
            self.logger.info(f"joining in feature set: {fs}")
            self.featureset_map[fs.name] = fs.to_proto()

        if not self.config.feature_sets:
            self.logger.info(
                f"No feature sets to join in for batch prediction: {batch_prediction_name}"
            )

    def run(self):
        """
        Constructs pipeline
        Constructs train args
        Constructs promote args
        Runs Train
        Runs Promote
        Logs errors
        """
        # Collect data
        try:
            predict_data = Dataset(
                model=self.model,
                featureset_map=self.featureset_map,
                data_store=self.data_store,
            )
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors("Error getting data", e, stack_trace)
            raise

        try:
            pipeline: BaseModelPipeline = get_pipeline(
                client=self.client, pipeline_id=self.model.schema.pipeline.name
            )()
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors(f"Exception getting pipeline", e, stack_trace)
            raise

        # Do batch prediction
        self._batch_predict_args = dict(
            client=self.client,
            model_version_name=self.model_version_name,
            batch_prediction_name=self.batch_prediction_name,
            predict_data=predict_data,
            model=self.model,
            data_store=self.data_store,
            model_version_signature=self.model_version_signature,
            disable_feature_timestamp_generation=self.disable_feature_timestamp_generation,
            log_level=self.logging_level,
        )

        try:
            pipeline.predict(**self._batch_predict_args)
        except Exception as e:
            stack_trace = traceback.format_exc()
            self._set_errors(f"Exception during batchpredict", e, stack_trace)
            raise

        # Successful predict
        self._set_errors(clear=True)

    def _set_errors(
        self, message=None, e: Exception = None, stack_trace: str = None, clear=False
    ):
        """
        Make sure errors are recorded in the logs as well as the
        model version

        Parameters
        ---------
        message: str
            Description of what pipeline step the exception occured in
        e: Exception
            The exception itself
        """

        if clear:
            self.client.logger.log_field("error_message", "")
            self.client.logger.log_field("stack_trace", "")
        else:
            error_message = f"{message}:\n {e}"
            self.logger.error(f"{message}:\n {e} : {stack_trace}")
            self.client.logger.log_field("error_message", error_message)
            self.client.logger.log_field("stack_trace", stack_trace)
